"""Test index_list class"""

# pylint: disable=C0115, C0116, C0302, W0201, W0212
from copy import deepcopy
from unittest import TestCase
from unittest.mock import Mock
import yaml
from es_client.exceptions import FailedValidation
from curator.exceptions import (
    ActionError,
    ConfigurationError,
    FailedExecution,
    MissingArgument,
    NoIndices,
)
from curator.helpers.date_ops import fix_epoch
from curator import IndexList

# Get test variables and constants from a single source
from . import testvars


def get_es_ver():
    return {'version': {'number': '8.0.0'}}


def get_testvals(number, key):
    """Return the appropriate value per the provided key number"""
    data = {
        "1": {
            "settings": testvars.settings_one,
            "state": testvars.state_one,
            "stats": testvars.stats_one,
            "fieldstats": testvars.fieldstats_one,
        },
        "2": {
            "settings": testvars.settings_two,
            "state": testvars.state_two,
            "stats": testvars.stats_two,
            "fieldstats": testvars.fieldstats_two,
        },
        "4": {
            "settings": testvars.settings_four,
            "state": testvars.state_four,
            "stats": testvars.stats_four,
            "fieldstats": testvars.fieldstats_four,
        },
    }
    return data[number][key]


class TestIndexListClientAndInit(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_init_bad_client(self):
        client = 'not a real client'
        self.assertRaises(TypeError, IndexList, client)

    def test_init_get_indices_exception(self):
        self.builder()
        self.client.cat.indices.side_effect = testvars.fake_fail
        self.assertRaises(FailedExecution, IndexList, self.client)

    def test_init(self):
        self.builder()
        self.ilo.get_index_stats()
        self.ilo.get_index_settings()
        self.assertEqual(
            testvars.stats_two['indices']['index-2016.03.03']['total']['store'][
                'size_in_bytes'
            ],
            self.ilo.index_info['index-2016.03.03']['size_in_bytes'],
        )
        self.assertEqual(
            testvars.state_two[1]['status'],
            self.ilo.index_info['index-2016.03.04']['state'],
        )
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_for_closed_index(self):
        self.builder()
        self.client.cat.indices.return_value = testvars.state_2_closed
        self.client.indices.get_settings.return_value = testvars.settings_2_closed
        ilo2 = IndexList(self.client)
        ilo2.get_index_state()
        self.assertEqual('close', ilo2.index_info['index-2016.03.03']['state'])


class TestIndexListOtherMethods(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_empty_list(self):
        self.builder()
        self.client.indices.exists_alias.return_value = False
        self.assertEqual(2, len(self.ilo.indices))
        self.ilo.indices = []
        self.assertRaises(NoIndices, self.ilo.empty_list_check)

    def test_get_segmentcount(self):
        self.builder(key='1')
        self.client.indices.segments.return_value = testvars.shards
        # Ordinarily get_index_state is run before get_segment_counts, so we do
        # so manually here.
        self.ilo.get_index_state()
        self.ilo.get_segment_counts()
        self.assertEqual(71, self.ilo.index_info[testvars.named_index]['segments'])


class TestIndexListAgeFilterName(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_get_name_based_ages_match(self):
        self.builder()
        self.ilo.get_index_state()
        self.ilo.get_index_settings()
        self.ilo._get_name_based_ages('%Y.%m.%d')
        self.assertEqual(
            1456963200, self.ilo.index_info['index-2016.03.03']['age']['name']
        )

    def test_get_name_based_ages_no_match(self):
        self.builder()
        self.ilo.get_index_settings()
        self.ilo._get_name_based_ages('%Y-%m-%d')
        self.assertEqual(
            fix_epoch(
                testvars.settings_two['index-2016.03.03']['settings']['index'][
                    'creation_date'
                ]
            ),
            self.ilo.index_info['index-2016.03.03']['age']['creation_date'],
        )


class TestIndexListAgeFilterStatsAPI(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_get_field_stats_dates_negative(self):
        self.builder()
        self.client.search.return_value = testvars.fieldstats_query
        self.client.field_stats.return_value = testvars.fieldstats_two
        self.ilo._get_field_stats_dates(field='timestamp')
        self.assertNotIn('not_an_index_name', list(self.ilo.index_info.keys()))

    def test_get_field_stats_dates_field_not_found(self):
        self.builder()
        self.client.search.return_value = {'aggregations': {'foo': 'bar'}}
        self.assertRaises(
            ActionError, self.ilo._get_field_stats_dates, field='not_in_index'
        )


class TestIndexListRegexFilters(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_filter_by_regex_prefix(self):
        self.builder()
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='prefix', value='ind')
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='prefix', value='ind', exclude=True)
        self.assertEqual([], self.ilo.indices)

    def test_filter_by_regex_middle(self):
        self.builder()
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='regex', value='dex')
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='regex', value='dex', exclude=True)
        self.assertEqual([], self.ilo.indices)

    def test_filter_by_regex_timestring(self):
        self.builder()
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='timestring', value='%Y.%m.%d')
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='timestring', value='%Y.%m.%d', exclude=True)
        self.assertEqual([], self.ilo.indices)

    def test_filter_by_regex_no_match_exclude(self):
        self.builder()
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='prefix', value='invalid', exclude=True)
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_filter_by_regex_no_value(self):
        self.builder()
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.assertRaises(
            ValueError, self.ilo.filter_by_regex, kind='prefix', value=None
        )
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.ilo.filter_by_regex(kind='prefix', value=0)
        self.assertEqual([], self.ilo.indices)

    def test_filter_by_regex_bad_kind(self):
        self.builder()
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )
        self.assertRaises(
            ValueError, self.ilo.filter_by_regex, kind='invalid', value=None
        )


class TestIndexListFilterByAge(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_missing_direction(self):
        self.builder()
        self.assertRaises(
            MissingArgument, self.ilo.filter_by_age, unit='days', unit_count=1
        )

    def test_bad_direction(self):
        self.builder()
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_age,
            unit='days',
            unit_count=1,
            direction="invalid",
        )

    def test_name_no_timestring(self):
        self.builder()
        self.assertRaises(
            MissingArgument,
            self.ilo.filter_by_age,
            source='name',
            unit='days',
            unit_count=1,
            direction='older',
        )

    def test_name_older_than_now(self):
        self.builder()
        self.ilo.filter_by_age(
            source='name',
            direction='older',
            timestring='%Y.%m.%d',
            unit='days',
            unit_count=1,
        )
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_name_older_than_now_exclude(self):
        self.builder()
        self.ilo.filter_by_age(
            source='name',
            direction='older',
            timestring='%Y.%m.%d',
            unit='days',
            unit_count=1,
            exclude=True,
        )
        self.assertEqual([], sorted(self.ilo.indices))

    def test_name_younger_than_now(self):
        self.builder()
        self.ilo.filter_by_age(
            source='name',
            direction='younger',
            timestring='%Y.%m.%d',
            unit='days',
            unit_count=1,
        )
        self.assertEqual([], sorted(self.ilo.indices))

    def test_name_younger_than_now_exclude(self):
        self.builder()
        self.ilo.filter_by_age(
            source='name',
            direction='younger',
            timestring='%Y.%m.%d',
            unit='days',
            unit_count=1,
            exclude=True,
        )
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_name_younger_than_past_date(self):
        self.builder()
        self.ilo.filter_by_age(
            source='name',
            direction='younger',
            timestring='%Y.%m.%d',
            unit='seconds',
            unit_count=0,
            epoch=1457049599,
        )
        self.assertEqual(['index-2016.03.04'], sorted(self.ilo.indices))

    def test_name_older_than_past_date(self):
        self.builder()
        self.ilo.filter_by_age(
            source='name',
            direction='older',
            timestring='%Y.%m.%d',
            unit='seconds',
            unit_count=0,
            epoch=1456963201,
        )
        self.assertEqual(['index-2016.03.03'], sorted(self.ilo.indices))

    def test_creation_date_older_than_now(self):
        self.builder()
        self.ilo.filter_by_age(
            source='creation_date', direction='older', unit='days', unit_count=1
        )
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_creation_date_older_than_now_raises(self):
        self.builder()
        self.ilo.get_index_state()
        self.ilo.get_index_settings()
        self.ilo.index_info['index-2016.03.03']['age'].pop('creation_date')
        self.ilo.index_info['index-2016.03.04']['age'].pop('creation_date')
        self.ilo.filter_by_age(
            source='creation_date', direction='older', unit='days', unit_count=1
        )
        self.assertEqual([], self.ilo.indices)

    def test_creation_date_younger_than_now(self):
        self.builder()
        self.ilo.filter_by_age(
            source='creation_date', direction='younger', unit='days', unit_count=1
        )
        self.assertEqual([], sorted(self.ilo.indices))

    def test_creation_date_younger_than_now_raises(self):
        self.builder()
        self.ilo.get_index_state()
        self.ilo.get_index_settings()
        self.ilo.index_info['index-2016.03.03']['age'].pop('creation_date')
        self.ilo.index_info['index-2016.03.04']['age'].pop('creation_date')
        self.ilo.filter_by_age(
            source='creation_date', direction='younger', unit='days', unit_count=1
        )
        self.assertEqual([], self.ilo.indices)

    def test_creation_date_younger_than_past_date(self):
        self.builder()
        self.ilo.filter_by_age(
            source='creation_date',
            direction='younger',
            unit='seconds',
            unit_count=0,
            epoch=1457049599,
        )
        self.assertEqual(['index-2016.03.04'], sorted(self.ilo.indices))

    def test_creation_date_older_than_past_date(self):
        self.builder()
        self.ilo.filter_by_age(
            source='creation_date',
            direction='older',
            unit='seconds',
            unit_count=0,
            epoch=1456963201,
        )
        self.assertEqual(['index-2016.03.03'], sorted(self.ilo.indices))

    def test_field_stats_missing_field(self):
        self.builder()
        self.assertRaises(
            MissingArgument,
            self.ilo.filter_by_age,
            source='field_stats',
            direction='older',
            unit='days',
            unit_count=1,
        )

    def test_field_stats_invalid_stats_result(self):
        self.builder()
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_age,
            field='timestamp',
            source='field_stats',
            direction='older',
            unit='days',
            unit_count=1,
            stats_result='invalid',
        )

    def test_field_stats_invalid_source(self):
        self.builder()
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_age,
            source='invalid',
            direction='older',
            unit='days',
            unit_count=1,
        )


class TestIndexListFilterBySpace(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.search.return_value = get_testvals(key, 'fieldstats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_missing_disk_space_value(self):
        self.builder()
        self.assertRaises(MissingArgument, self.ilo.filter_by_space)

    def test_filter_result_by_name(self):
        self.builder()
        self.ilo.filter_by_space(disk_space=1.1)
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_filter_result_by_name_reverse_order(self):
        self.builder()
        self.ilo.filter_by_space(disk_space=1.1, reverse=False)
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_filter_result_by_name_exclude(self):
        self.builder()
        self.ilo.filter_by_space(disk_space=1.1, exclude=True)
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_filter_result_by_date_raise(self):
        self.builder(key='4')
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_space,
            disk_space=2.1,
            use_age=True,
            source='invalid',
        )

    def test_filter_result_by_date_timestring_raise(self):
        self.builder(key='4')
        self.assertRaises(
            MissingArgument,
            self.ilo.filter_by_space,
            disk_space=2.1,
            use_age=True,
            source='name',
        )

    def test_filter_result_by_date_timestring(self):
        self.builder(key='4')
        self.ilo.filter_by_space(
            disk_space=2.1, use_age=True, source='name', timestring='%Y.%m.%d'
        )
        self.assertEqual(['a-2016.03.03'], sorted(self.ilo.indices))

    def test_filter_result_by_date_non_matching_timestring(self):
        self.builder(key='4')
        self.ilo.filter_by_space(
            disk_space=2.1, use_age=True, source='name', timestring='%Y.%m.%d.%H'
        )
        self.assertEqual([], sorted(self.ilo.indices))

    def test_filter_threshold_behavior(self):
        self.builder()
        il_a = IndexList(self.client)
        # less than
        il_a.filter_by_space(
            disk_space=1.5, use_age=True, threshold_behavior='less_than'
        )
        self.assertEqual(['index-2016.03.04'], sorted(il_a.indices))
        # greater than
        il_b = IndexList(self.client)
        il_b.filter_by_space(
            disk_space=1.5, use_age=True, threshold_behavior='greater_than'
        )
        self.assertEqual(['index-2016.03.03'], sorted(il_b.indices))
        # default case
        il_c = IndexList(self.client)
        il_c.filter_by_space(disk_space=1.5, use_age=True)
        self.assertEqual(['index-2016.03.03'], sorted(il_c.indices))

    def test_filter_bad_threshold_behavior(self):
        self.builder()
        # less than
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_space,
            disk_space=1.5,
            threshold_behavior='invalid',
        )

    def test_filter_result_by_date_field_stats_raise(self):
        self.builder(key='4')
        self.client.search.return_value = testvars.fieldstats_query
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_space,
            disk_space=2.1,
            use_age=True,
            source='min_value',
        )

    def test_filter_result_by_date_no_field_raise(self):
        self.builder(key='4')
        self.assertRaises(
            MissingArgument,
            self.ilo.filter_by_space,
            disk_space=2.1,
            use_age=True,
            source='field_stats',
        )

    def test_filter_result_by_date_invalid_stats_result_raise(self):
        self.builder(key='4')
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_space,
            disk_space=2.1,
            use_age=True,
            source='field_stats',
            field='timestamp',
            stats_result='invalid',
        )

    def test_filter_result_by_creation_date(self):
        self.builder(key='4')
        self.ilo.filter_by_space(disk_space=2.1, use_age=True)
        self.assertEqual(['a-2016.03.03'], self.ilo.indices)


class TestIndexListFilterKibana(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.search.return_value = get_testvals(key, 'fieldstats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_filter_kibana_positive(self):
        self.builder()
        # Establish the object per requirements, then overwrite
        self.ilo.indices = ['.kibana', '.kibana-5', '.kibana-6', 'dummy']
        self.ilo.filter_kibana()
        self.assertEqual(['dummy'], self.ilo.indices)

    def test_filter_kibana_positive_include(self):
        self.builder()
        # Establish the object per requirements, then overwrite
        self.ilo.indices = ['.kibana', '.kibana-5', '.kibana-6', 'dummy']
        self.ilo.filter_kibana(exclude=False)
        self.assertEqual(['.kibana', '.kibana-5', '.kibana-6'], self.ilo.indices)

    def test_filter_kibana_positive_exclude(self):
        self.builder()
        # Establish the object per requirements, then overwrite
        kibana_indices = ['.kibana', '.kibana-5', '.kibana-6']
        self.ilo.indices = kibana_indices
        self.ilo.indices.append('dummy')
        self.ilo.filter_kibana(exclude=True)
        self.assertEqual(kibana_indices, self.ilo.indices)

    def test_filter_kibana_negative(self):
        self.builder()
        # Establish the object per requirements, then overwrite
        self.ilo.indices = [
            'kibana',
            'marvel-kibana',
            'cabana-int',
            'marvel-es-data',
            'dummy',
        ]
        self.ilo.filter_kibana()
        self.assertEqual(
            ['kibana', 'marvel-kibana', 'cabana-int', 'marvel-es-data', 'dummy'],
            self.ilo.indices,
        )


class TestIndexListFilterForceMerged(TestCase):
    VERSION = {'version': {'number': '8.0.0'}}

    def builder(self):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.segments.return_value = testvars.shards
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_filter_forcemerge_raise(self):
        self.builder()
        self.assertRaises(MissingArgument, self.ilo.filter_forceMerged)

    def test_filter_forcemerge_positive(self):
        self.builder()
        self.ilo.filter_forceMerged(max_num_segments=2)
        self.assertEqual([testvars.named_index], self.ilo.indices)

    def test_filter_forcemerge_negative(self):
        self.builder()
        self.client.indices.segments.return_value = testvars.fm_shards
        self.ilo.filter_forceMerged(max_num_segments=2)
        self.assertEqual([], self.ilo.indices)


class TestIndexListFilterOpened(TestCase):
    def test_filter_opened(self):
        client = Mock()
        client.info.return_value = {'version': {'number': '8.0.0'}}
        client.cat.indices.return_value = testvars.state_four
        client.indices.get_settings.return_value = testvars.settings_four
        client.indices.stats.return_value = testvars.stats_four
        client.field_stats.return_value = testvars.fieldstats_four
        client.indices.exists_alias.return_value = False
        ilo = IndexList(client)
        ilo.filter_opened()
        self.assertEqual(['c-2016.03.05'], ilo.indices)


class TestIndexListFilterAllocated(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_missing_key(self):
        self.builder()
        self.assertRaises(
            MissingArgument,
            self.ilo.filter_allocated,
            value='foo',
            allocation_type='invalid',
        )

    def test_missing_value(self):
        self.builder()
        self.assertRaises(
            MissingArgument,
            self.ilo.filter_allocated,
            key='tag',
            allocation_type='invalid',
        )

    def test_invalid_allocation_type(self):
        self.builder()
        self.assertRaises(
            ValueError,
            self.ilo.filter_allocated,
            key='tag',
            value='foo',
            allocation_type='invalid',
        )

    def test_success(self):
        self.builder()
        self.ilo.filter_allocated(key='tag', value='foo', allocation_type='include')
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_invalid_tag(self):
        self.builder()
        self.ilo.filter_allocated(key='invalid', value='foo', allocation_type='include')
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )


class TestIterateFiltersIndex(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_no_filters(self):
        self.builder(key='4')
        self.ilo.iterate_filters({})
        self.assertEqual(
            ['a-2016.03.03', 'b-2016.03.04', 'c-2016.03.05', 'd-2016.03.06'],
            sorted(self.ilo.indices),
        )

    def test_no_filtertype(self):
        self.builder(key='4')
        config = {'filters': [{'no_filtertype': 'fail'}]}
        self.assertRaises(FailedValidation, self.ilo.iterate_filters, config)

    def test_invalid_filtertype(self):
        self.builder(key='4')
        config = {'filters': [{'filtertype': 12345.6789}]}
        self.assertRaises(FailedValidation, self.ilo.iterate_filters, config)

    def test_pattern_filtertype(self):
        self.builder(key='4')
        config = yaml.load(testvars.pattern_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['a-2016.03.03'], self.ilo.indices)

    def test_age_filtertype(self):
        self.builder()
        config = yaml.load(testvars.age_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_space_filtertype(self):
        self.builder(key='4')
        self.client.field_stats.return_value = testvars.fieldstats_four
        config = yaml.load(testvars.space_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['a-2016.03.03'], self.ilo.indices)

    def test_forcemerge_filtertype(self):
        self.builder(key='1')
        self.client.indices.segments.return_value = testvars.shards
        config = yaml.load(testvars.forcemerge_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual([testvars.named_index], self.ilo.indices)

    def test_allocated_filtertype(self):
        self.builder()
        config = yaml.load(testvars.allocated_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_kibana_filtertype(self):
        self.builder()
        self.client.field_stats.return_value = testvars.fieldstats_two
        # Establish the object per requirements, then overwrite
        self.ilo.indices = ['.kibana', '.kibana-5', '.kibana-6', 'dummy']
        config = yaml.load(testvars.kibana_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['dummy'], self.ilo.indices)

    def test_opened_filtertype(self):
        self.builder(key='4')
        self.client.field_stats.return_value = testvars.fieldstats_four
        config = yaml.load(testvars.opened_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['c-2016.03.05'], self.ilo.indices)

    def test_closed_filtertype(self):
        self.builder(key='4')
        self.client.field_stats.return_value = testvars.fieldstats_four
        config = yaml.load(testvars.closed_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(
            ['a-2016.03.03', 'b-2016.03.04', 'd-2016.03.06'], sorted(self.ilo.indices)
        )

    def test_none_filtertype(self):
        self.builder()
        config = yaml.load(testvars.none_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_unknown_filtertype_raises(self):
        self.builder()
        config = yaml.load(testvars.invalid_ft, Loader=yaml.FullLoader)['actions'][1]
        self.assertRaises(FailedValidation, self.ilo.iterate_filters, config)

    def test_ilm_filtertype_exclude(self):
        self.builder()
        # If we don't deepcopy, then it munges the settings for future references.
        with_ilm = deepcopy(testvars.settings_two)
        with_ilm['index-2016.03.03']['settings']['index']['lifecycle'] = {
            'name': 'mypolicy'
        }
        self.client.indices.get_settings.return_value = with_ilm
        config = {'filters': [{'filtertype': 'ilm', 'exclude': True}]}
        self.ilo.iterate_filters(config)
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_ilm_filtertype_no_setting(self):
        self.builder()
        config = {'filters': [{'filtertype': 'ilm', 'exclude': True}]}
        self.ilo.iterate_filters(config)
        self.assertEqual(
            ['index-2016.03.03', 'index-2016.03.04'], sorted(self.ilo.indices)
        )

    def test_size_filtertype(self):
        self.builder()
        config = yaml.load(testvars.size_ft, Loader=yaml.FullLoader)['actions'][1]
        self.ilo.iterate_filters(config)
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)


class TestIndexListFilterAlias(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_raise(self):
        self.builder(key='1')
        self.assertRaises(MissingArgument, self.ilo.filter_by_alias)

    def test_positive(self):
        self.builder()
        self.client.indices.get_alias.return_value = testvars.settings_2_get_aliases
        self.ilo.filter_by_alias(aliases=['my_alias'])
        self.assertEqual(
            sorted(list(testvars.settings_two.keys())), sorted(self.ilo.indices)
        )

    def test_negative(self):
        self.builder()
        self.client.indices.get_alias.return_value = {}
        self.ilo.filter_by_alias(aliases=['not_my_alias'])
        self.assertEqual(sorted([]), sorted(self.ilo.indices))

    def test_get_alias_raises(self):
        self.builder()
        self.client.indices.get_alias.side_effect = testvars.get_alias_fail
        self.client.indices.get_alias.return_value = testvars.settings_2_get_aliases
        self.ilo.filter_by_alias(aliases=['my_alias'])
        self.assertEqual(sorted([]), sorted(self.ilo.indices))


class TestIndexListFilterCount(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.client.indices.get_alias.return_value = testvars.settings_2_get_aliases
        self.ilo = IndexList(self.client)

    def test_raise(self):
        self.builder()
        self.assertRaises(MissingArgument, self.ilo.filter_by_count)

    def test_without_age(self):
        self.builder()
        self.ilo.filter_by_count(count=1)
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_without_age_reversed(self):
        self.builder()
        self.ilo.filter_by_count(count=1, reverse=False)
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_with_age(self):
        self.builder()
        self.ilo.filter_by_count(
            count=1, use_age=True, source='name', timestring='%Y.%m.%d'
        )
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_with_age_creation_date(self):
        self.builder()
        self.ilo.filter_by_count(count=1, use_age=True)
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_with_age_reversed(self):
        self.builder()
        self.ilo.filter_by_count(
            count=1, use_age=True, source='name', timestring='%Y.%m.%d', reverse=False
        )
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_pattern_no_regex_group(self):
        self.builder()
        self.assertRaises(
            ActionError,
            self.ilo.filter_by_count,
            count=1,
            use_age=True,
            pattern=' ',
            source='name',
            timestring='%Y.%m.%d',
        )

    def test_pattern_multiple_regex_groups(self):
        self.builder()
        self.assertRaises(
            ActionError,
            self.ilo.filter_by_count,
            count=1,
            use_age=True,
            pattern=r'^(\ )foo(\ )$',
            source='name',
            timestring='%Y.%m.%d',
        )


class TestIndexListFilterShards(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_filter_shards_raise(self):
        self.builder()
        self.assertRaises(MissingArgument, self.ilo.filter_by_shards)

    def test_bad_shard_count_raise_1(self):
        self.builder()
        self.assertRaises(
            MissingArgument, self.ilo.filter_by_shards, number_of_shards=0
        )

    def test_bad_shard_count_raise_2(self):
        self.builder()
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_shards,
            number_of_shards=1,
            shard_filter_behavior='less_than',
        )

    def test_bad_shard_count_raise_3(self):
        self.builder()
        self.assertRaises(
            ValueError,
            self.ilo.filter_by_shards,
            number_of_shards=-1,
            shard_filter_behavior='greater_than',
        )

    def test_greater_than_or_equal(self):
        self.builder()
        self.ilo.filter_by_shards(
            number_of_shards=5, shard_filter_behavior='greater_than_or_equal'
        )
        self.assertEqual(
            sorted(['index-2016.03.03', 'index-2016.03.04']), sorted(self.ilo.indices)
        )

    def test_greater_than_or_equal_exclude(self):
        self.builder()
        self.ilo.filter_by_shards(
            number_of_shards=5,
            shard_filter_behavior='greater_than_or_equal',
            exclude=True,
        )
        self.assertEqual(sorted([]), sorted(self.ilo.indices))

    def test_greater_than(self):
        self.builder()
        self.ilo.filter_by_shards(number_of_shards=5)
        self.assertEqual(sorted([]), sorted(self.ilo.indices))

    def test_greater_than_exclude(self):
        self.builder()
        self.ilo.filter_by_shards(number_of_shards=5, exclude=True)
        self.assertEqual(
            sorted(['index-2016.03.03', 'index-2016.03.04']), sorted(self.ilo.indices)
        )

    def test_less_than_or_equal(self):
        self.builder()
        self.ilo.filter_by_shards(
            number_of_shards=5, shard_filter_behavior='less_than_or_equal'
        )
        self.assertEqual(
            sorted(['index-2016.03.03', 'index-2016.03.04']), sorted(self.ilo.indices)
        )

    def test_less_than_or_equal_exclude(self):
        self.builder()
        self.ilo.filter_by_shards(
            number_of_shards=5, shard_filter_behavior='less_than_or_equal', exclude=True
        )
        self.assertEqual(sorted([]), sorted(self.ilo.indices))

    def test_less_than(self):
        self.builder()
        self.ilo.filter_by_shards(number_of_shards=5, shard_filter_behavior='less_than')
        self.assertEqual(sorted([]), sorted(self.ilo.indices))

    def test_less_than_exclude(self):
        self.builder()
        self.ilo.filter_by_shards(
            number_of_shards=5, shard_filter_behavior='less_than', exclude=True
        )
        self.assertEqual(
            sorted(['index-2016.03.03', 'index-2016.03.04']), sorted(self.ilo.indices)
        )

    def test_equal(self):
        self.builder()
        self.ilo.filter_by_shards(number_of_shards=5, shard_filter_behavior='equal')
        self.assertEqual(
            sorted(['index-2016.03.03', 'index-2016.03.04']), sorted(self.ilo.indices)
        )


class TestIndexListPeriodFilterName(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)
        self.timestring = '%Y.%m.%d'
        self.epoch = 1456963201
        self.unit = 'days'

    def test_get_name_based_age_in_range(self):
        range_from = -1
        range_to = 0
        expected = ['index-2016.03.03']
        self.builder()
        self.ilo.filter_period(
            unit=self.unit,
            range_from=range_from,
            range_to=range_to,
            source='name',
            timestring=self.timestring,
            epoch=self.epoch,
        )
        self.assertEqual(expected, self.ilo.indices)

    def test_get_name_based_age_not_in_range(self):
        range_from = -3
        range_to = -2
        expected = []
        self.builder()
        self.ilo.filter_period(
            unit=self.unit,
            range_from=range_from,
            range_to=range_to,
            source='name',
            timestring=self.timestring,
            epoch=self.epoch,
        )
        self.assertEqual(expected, self.ilo.indices)

    def test_bad_arguments(self):
        range_from = -2
        range_to = -3
        self.builder()
        self.assertRaises(
            FailedExecution,
            self.ilo.filter_period,
            unit=self.unit,
            range_from=range_from,
            range_to=range_to,
            source='name',
            timestring=self.timestring,
            epoch=self.epoch,
        )

    def test_missing_creation_date_raises(self):
        range_from = -1
        range_to = 0
        expected = []
        self.builder()
        self.ilo.get_index_state()
        self.ilo.get_index_settings()
        self.ilo.index_info['index-2016.03.03']['age'].pop('creation_date')
        self.ilo.index_info['index-2016.03.04']['age'].pop('creation_date')
        self.ilo.filter_period(
            unit=self.unit,
            range_from=range_from,
            range_to=range_to,
            source='creation_date',
            epoch=self.epoch,
        )
        self.assertEqual(expected, self.ilo.indices)

    def test_non_integer_range_value(self):
        self.builder()
        self.assertRaises(
            ConfigurationError, self.ilo.filter_period, range_from='invalid'
        )


class TestPeriodFilterAbsolute(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_bad_period_type(self):
        self.builder()
        self.assertRaises(ValueError, self.ilo.filter_period, period_type='invalid')

    def test_none_value_raises(self):
        self.builder()
        self.assertRaises(
            ConfigurationError,
            self.ilo.filter_period,
            period_type='absolute',
            date_from=None,
        )

    def test_fail_on_bad_date(self):
        unit = 'months'
        date_from = '2016.17'
        date_from_format = '%Y.%m'
        date_to = '2017.01'
        date_to_format = '%Y.%m'
        self.builder()
        self.assertRaises(
            FailedExecution,
            self.ilo.filter_period,
            unit=unit,
            source='creation_date',
            period_type='absolute',
            date_from=date_from,
            date_to=date_to,
            date_from_format=date_from_format,
            date_to_format=date_to_format,
        )


class TestIndexListFilterBySize(TestCase):
    def builder(self, key='2'):
        self.client = Mock()
        self.client.info.return_value = get_es_ver()
        self.client.cat.indices.return_value = get_testvals(key, 'state')
        self.client.indices.get_settings.return_value = get_testvals(key, 'settings')
        self.client.indices.stats.return_value = get_testvals(key, 'stats')
        self.client.field_stats.return_value = get_testvals(key, 'fieldstats')
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_missing_size_value(self):
        self.builder()
        self.assertRaises(MissingArgument, self.ilo.filter_by_size)

    def test_filter_default_result(self):
        self.builder()
        self.ilo.filter_by_size(size_threshold=0.52)
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_filter_default_result_and_exclude(self):
        self.builder()
        self.ilo.filter_by_size(size_threshold=0.52, exclude=True)
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_filter_by_threshold_behavior_less_than(self):
        self.builder()
        self.ilo.filter_by_size(size_threshold=0.52, threshold_behavior='less_than')
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)

    def test_filter_by_size_behavior_total(self):
        self.builder()
        self.ilo.filter_by_size(size_threshold=1.04, size_behavior='total')
        self.assertEqual(['index-2016.03.04'], self.ilo.indices)

    def test_filter_by_size_behavior_total_and_threshold_behavior_less_than(self):
        self.builder()
        self.ilo.filter_by_size(
            size_threshold=1.04, size_behavior='total', threshold_behavior='less_than'
        )
        self.assertEqual(['index-2016.03.03'], self.ilo.indices)



"""Test delete snapshot functionality"""

# pylint: disable=C0115, C0116, invalid-name
import os
import time
from curator import IndexList
from curator.actions.snapshot import Snapshot
from curator.helpers.getters import get_snapshot
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')
# '      repository: {0}\n'
# '      - filtertype: {1}\n'
# '        source: {2}\n'
# '        direction: {3}\n'
# '        timestring: {4}\n'
# '        unit: {5}\n'
# '        unit_count: {6}\n'
# '        epoch: {7}\n')


class TestActionFileDeleteSnapshots(CuratorTestCase):
    def test_deletesnapshot(self):
        # Create snapshots to delete and verify them
        self.create_repository()
        timestamps = []
        for i in range(1, 4):
            self.add_docs(f'my_index{i}')
            ilo = IndexList(self.client)
            snap = Snapshot(
                ilo,
                repository=self.args['repository'],
                name='curator-%Y%m%d%H%M%S',
                wait_interval=0.5,
            )
            snap.do_action()
            snapshot = get_snapshot(self.client, self.args['repository'], '*')
            assert i == len(snapshot['snapshots'])
            time.sleep(1.0)
            timestamps.append(int(time.time()))
            time.sleep(1.0)
        # Setup the actual delete
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_snap_proto.format(
                self.args['repository'],
                'age',
                'creation_date',
                'older',
                ' ',
                'seconds',
                '0',
                timestamps[0],
            ),
        )
        self.invoke_runner()
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        assert 2 == len(snapshot['snapshots'])

    def test_no_repository(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.delete_snap_proto.format(
                ' ', 'age', 'creation_date', 'older', ' ', 'seconds', '0', ' '
            ),
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_extra_options(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.bad_option_proto_test.format('delete_snapshots'),
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code


class TestCLIDeleteSnapshots(CuratorTestCase):
    def test_deletesnapshot(self):
        # Create snapshots to delete and verify them
        self.create_repository()
        timestamps = []
        for i in range(1, 4):
            self.add_docs(f'my_index{i}')
            ilo = IndexList(self.client)
            snap = Snapshot(
                ilo,
                repository=self.args['repository'],
                name='curator-%Y%m%d%H%M%S',
                wait_interval=0.5,
            )
            snap.do_action()
            snapshot = get_snapshot(self.client, self.args['repository'], '*')
            assert i == len(snapshot['snapshots'])
            time.sleep(1.0)
            timestamps.append(int(time.time()))
            time.sleep(1.0)
        filter_list = (
            '{"filtertype":"age","source":"creation_date","direction":"older",'
            '"unit":"seconds","unit_count":0,"epoch":' + str(timestamps[0]) + '}'
        )
        # Setup the actual delete
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'delete-snapshots',
            '--repository',
            self.args['repository'],
            '--filter_list',
            filter_list,
        ]
        assert 0 == self.run_subprocess(
            args, logname='TestCLIDeleteSnapshots.test_deletesnapshot'
        )
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        assert 2 == len(snapshot['snapshots'])

    def test_count_by_age(self):
        self.create_repository()
        timestamps = []

        def add_snap(num, name):
            self.add_docs(f'my_index{num}')
            ilo = IndexList(self.client)
            snap = Snapshot(
                ilo, repository=self.args['repository'], name=name, wait_interval=0.5
            )
            snap.do_action()
            snapshot = get_snapshot(self.client, self.args['repository'], '*')
            assert num == len(snapshot['snapshots'])
            time.sleep(1.0)
            timestamps.append(int(time.time()))
            time.sleep(1.0)

        name = 'curator-%Y%m%d%H%M%S'
        for i in range(1, 4):
            add_snap(i, name)
        add_snap(4, 'kibana-index')
        # Setup the actual delete
        filter_list = (
            '{"filtertype":"count","count":2,"use_age":true,"source":"name",'
            '"timestring":"%Y%m%d%H%M%S"}'
        )
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'delete-snapshots',
            '--repository',
            self.args['repository'],
            '--filter_list',
            filter_list,
        ]
        assert 0 == self.run_subprocess(
            args, logname='TestCLIDeleteSnapshots.test_deletesnapshot'
        )
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        assert 3 == len(snapshot['snapshots'])



"""test_action_clusterrouting"""
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import ClusterRouting
# Get test variables and constants from a single source
from . import testvars

class TestActionAllocation(TestCase):
    def test_bad_client(self):
        self.assertRaises(TypeError, ClusterRouting, 'invalid', setting='enable')
    def test_bad_setting(self):
        client = Mock()
        self.assertRaises(
            ValueError, ClusterRouting, client, setting='invalid'
        )
    def test_bad_routing_type(self):
        client = Mock()
        self.assertRaises(
            ValueError,
            ClusterRouting,
            client,
            routing_type='invalid',
            setting='enable'
        )
    def test_bad_value_with_allocation(self):
        client = Mock()
        self.assertRaises(
            ValueError,
            ClusterRouting,
            client,
            routing_type='allocation',
            setting='enable',
            value='invalid'
        )
    def test_bad_value_with_rebalance(self):
        client = Mock()
        self.assertRaises(
            ValueError,
            ClusterRouting,
            client,
            routing_type='rebalance',
            setting='enable',
            value='invalid'
        )
    def test_do_dry_run(self):
        client = Mock()
        cro = ClusterRouting(
            client,
            routing_type='allocation',
            setting='enable',
            value='all'
        )
        self.assertIsNone(cro.do_dry_run())
    def test_do_action_raise_on_put_settings(self):
        client = Mock()
        client.cluster.put_settings.return_value = None
        client.cluster.put_settings.side_effect = testvars.fake_fail
        cro = ClusterRouting(
            client,
            routing_type='allocation',
            setting='enable',
            value='all'
        )
        self.assertRaises(Exception, cro.do_action)
    def test_do_action_wait(self):
        client = Mock()
        client.cluster.put_settings.return_value = None
        client.cluster.health.return_value = {'relocating_shards':0}
        cro = ClusterRouting(
            client,
            routing_type='allocation',
            setting='enable',
            value='all',
            wait_for_completion=True
        )
        self.assertIsNone(cro.do_action())



"""Snapshot and Restore action classes"""

import logging
from curator.helpers.getters import get_alias_actions, get_tier_preference, meta_getter
from curator.helpers.testers import (
    has_lifecycle_name,
    is_idx_partial,
    verify_index_list,
)
from curator.helpers.utils import report_failure
from curator.exceptions import (
    CuratorException,
    FailedExecution,
    SearchableSnapshotException,
)


class Cold2Frozen:
    """Cold to Frozen Tier Searchable Snapshot Action Class

    For manually migrating snapshots not associated with ILM from the cold tier
    to the frozen tier.
    """

    DEFAULTS = {
        'index_settings': None,
        'ignore_index_settings': ['index.refresh_interval'],
        'wait_for_completion': True,
    }

    def __init__(self, ilo, **kwargs):
        """
        :param ilo: An IndexList Object
        :param index_settings: (Optional) Settings that should be added to the
            index when it is mounted. If not set, set the ``_tier_preference``
            to the tiers available, coldest first.
        :param ignore_index_settings: (Optional, array of strings) Names of settings
            that should be removed from the index when it is mounted.
        :param wait_for_completion: Wait for completion before returning.

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type index_settings: dict
        :type ignore_index_settings: list
        :type wait_for_completion: bool
        """
        self.loggit = logging.getLogger('curator.actions.cold2frozen')
        verify_index_list(ilo)
        # Check here and don't bother with the rest of this if there are no
        # indices in the index list.
        ilo.empty_list_check()

        #: The :py:class:`~.curator.indexlist.IndexList` object passed from
        #: param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: Object attribute that contains the :py:func:`~.curator.helpers.utils.to_csv`
        #: output of the indices in :py:attr:`index_list`.
        self.indices = ilo
        #: Object attribute that gets the value of ``index_settings``.
        self.index_settings = None
        #: Object attribute that gets the value of ``ignore_index_settings``.
        self.ignore_index_settings = None
        #: Object attribute that gets the value of param ``wait_for_completion``.
        self.wait_for_completion = None

        # Parse the kwargs into attributes
        self.assign_kwargs(**kwargs)

    def assign_kwargs(self, **kwargs):
        """
        Assign the kwargs to the attribute of the same name with the passed value
        or the default from DEFAULTS
        """
        # Handy little loop here only adds kwargs that exist in DEFAULTS, or the
        # default value. It ignores any non-relevant kwargs
        for key, value in self.DEFAULTS.items():
            if key in kwargs:
                setattr(self, key, kwargs[key])
            else:
                setattr(self, key, value)

    def action_generator(self):
        """Yield a dict for use in :py:meth:`do_action` and :py:meth:`do_dry_run`

        :returns: A generator object containing the settings necessary to migrate
            indices from cold to frozen
        :rtype: dict
        """
        for idx in self.index_list.indices:
            idx_settings = meta_getter(self.client, idx, get='settings')
            self.loggit.debug('Index %s has settings: %s', idx, idx_settings)
            if has_lifecycle_name(idx_settings):
                self.loggit.critical(
                    'Index %s is associated with an ILM policy and this action  '
                    'will never work on an index associated with an ILM policy',
                    idx,
                )
                raise CuratorException(f'Index {idx} is associated with an ILM policy')
            if is_idx_partial(idx_settings):
                self.loggit.critical('Index %s is already in the frozen tier', idx)
                raise SearchableSnapshotException('Index is already in frozen tier')

            snap = idx_settings['store']['snapshot']['snapshot_name']
            snap_idx = idx_settings['store']['snapshot']['index_name']
            repo = idx_settings['store']['snapshot']['repository_name']
            msg = (
                f'Index {idx} Snapshot name: {snap}, Snapshot index: {snap_idx}, '
                f'repo: {repo}'
            )
            self.loggit.debug(msg)

            aliases = meta_getter(self.client, idx, get='alias')

            renamed = f'partial-{idx}'

            if not self.index_settings:
                self.index_settings = {
                    "routing": {
                        "allocation": {
                            "include": {
                                "_tier_preference": get_tier_preference(self.client)
                            }
                        }
                    }
                }
            yield {
                'repository': repo,
                'snapshot': snap,
                'index': snap_idx,
                'renamed_index': renamed,
                'index_settings': self.index_settings,
                'ignore_index_settings': self.ignore_index_settings,
                'storage': 'shared_cache',
                'wait_for_completion': self.wait_for_completion,
                'aliases': aliases,
                'current_idx': idx,
            }

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        for kwargs in self.action_generator():
            aliases = kwargs.pop('aliases')
            current_idx = kwargs.pop('current_idx')
            msg = (
                f'DRY-RUN: cold2frozen: from snapshot {kwargs["snapshot"]} in '
                f'repository {kwargs["repository"]}, mount index {kwargs["index"]} '
                f'renamed as  {kwargs["renamed_index"]} with index settings: '
                f'{kwargs["index_settings"]}  and ignoring settings: '
                f'{kwargs["ignore_index_settings"]}. wait_for_completion: '
                f'{kwargs["wait_for_completion"]}. Restore aliases: {aliases}. '
                f'Current index name: {current_idx}'
            )
            self.loggit.info(msg)

    def mount_index(self, newidx, kwargs):
        """
        Call :py:meth:`~.elasticsearch.client.SearchableSnapshotsClient.mount`
        to mount the indices in :py:attr:`ilo` in the Frozen tier.
        """
        try:
            self.loggit.debug('Mounting new index %s in frozen tier...', newidx)
            self.client.searchable_snapshots.mount(**kwargs)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)

    def verify_mount(self, newidx):
        """
        Verify that newidx is a mounted index
        """
        self.loggit.debug('Verifying new index %s is mounted properly...', newidx)
        idx_settings = self.client.indices.get(index=newidx)[newidx]
        if is_idx_partial(idx_settings['settings']['index']):
            self.loggit.info('Index %s is mounted for frozen tier', newidx)
        else:
            report_failure(
                SearchableSnapshotException(
                    f'Index {newidx} not a mounted searchable snapshot'
                )
            )

    def update_aliases(self, current_idx, newidx, aliases):
        """
        Call :py:meth:`~.elasticsearch.client.IndicesClient.update_aliases` to
        update each new frozen index with the aliases from the old cold-tier index.

        Verify aliases look good.
        """
        alias_names = aliases.keys()
        if not alias_names:
            self.loggit.warning('No aliases associated with index %s', current_idx)
        else:
            self.loggit.debug('Transferring aliases to new index %s', newidx)
            self.client.indices.update_aliases(
                actions=get_alias_actions(current_idx, newidx, aliases)
            )
            verify = self.client.indices.get(index=newidx)[newidx]['aliases'].keys()
            if alias_names != verify:
                self.loggit.error(
                    'Alias names do not match! %s does not match: %s',
                    alias_names,
                    verify,
                )
                report_failure(
                    FailedExecution('Aliases failed to transfer to new index')
                )

    def cleanup(self, current_idx, newidx):
        """
        Call :py:meth:`~.elasticsearch.client.IndicesClient.delete` to delete
        the cold tier index.
        """
        self.loggit.debug('Deleting old index: %s', current_idx)
        try:
            self.client.indices.delete(index=current_idx)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)
        self.loggit.info(
            'Successfully migrated %s to the frozen tier as %s', current_idx, newidx
        )

    def do_action(self):
        """
        Do the actions outlined:
        Extract values from generated kwargs
        Mount
        Verify
        Update Aliases
        Cleanup
        """
        for kwargs in self.action_generator():
            aliases = kwargs.pop('aliases')
            current_idx = kwargs.pop('current_idx')
            newidx = kwargs['renamed_index']

            # Mount the index
            self.mount_index(newidx, kwargs)

            # Verify it's mounted as a partial now:
            self.verify_mount(newidx)

            # Update Aliases
            self.update_aliases(current_idx, newidx, aliases)

            # Clean up old index
            self.cleanup(current_idx, newidx)



"""test_action_replicas"""
# pylint: disable=missing-function-docstring, missing-class-docstring, protected-access, attribute-defined-outside-init
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import Replicas
from curator.exceptions import FailedExecution, MissingArgument
from curator import IndexList
# Get test variables and constants from a single source
from . import testvars

class TestActionReplicas(TestCase):
    VERSION = {'version': {'number': '8.0.0'} }
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = self.VERSION
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.client.indices.put_settings.return_value = None
        self.ilo = IndexList(self.client)
    def test_init_raise_bad_client(self):
        self.assertRaises(TypeError, Replicas, 'invalid', count=2)
    def test_init_raise_no_count(self):
        self.builder()
        self.assertRaises(MissingArgument, Replicas, self.ilo)
    def test_init(self):
        self.builder()
        rpo = Replicas(self.ilo, count=2)
        self.assertEqual(self.ilo, rpo.index_list)
        self.assertEqual(self.client, rpo.client)
    def test_do_dry_run(self):
        self.builder()
        rpo = Replicas(self.ilo, count=0)
        self.assertIsNone(rpo.do_dry_run())
    def test_do_action(self):
        self.builder()
        rpo = Replicas(self.ilo, count=0)
        self.assertIsNone(rpo.do_action())
    def test_do_action_wait(self):
        self.builder()
        self.client.cluster.health.return_value = {'status':'green'}
        rpo = Replicas(self.ilo, count=1, wait_for_completion=True)
        self.assertIsNone(rpo.do_action())
    def test_do_action_raises_exception(self):
        self.builder()
        self.client.indices.segments.return_value = testvars.shards
        self.client.indices.put_settings.side_effect = testvars.fake_fail
        rpo = Replicas(self.ilo, count=2)
        self.assertRaises(FailedExecution, rpo.do_action)



"""test_action_snapshot"""
# pylint: disable=missing-function-docstring, missing-class-docstring, line-too-long, protected-access, attribute-defined-outside-init
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import Snapshot
from curator.exceptions import ActionError, CuratorException, FailedExecution, FailedSnapshot, MissingArgument, SnapshotInProgress
from curator import IndexList
# Get test variables and constants from a single source
from . import testvars

class TestActionSnapshot(TestCase):
    VERSION = {'version': {'number': '8.0.0'} }
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = self.VERSION
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.client.snapshot.get_repository.return_value = testvars.test_repo
        self.client.snapshot.get.return_value = testvars.snapshots
        self.client.tasks.get.return_value = testvars.no_snap_tasks
        self.ilo = IndexList(self.client)
    def test_init_raise_bad_index_list(self):
        self.assertRaises(TypeError, Snapshot, 'invalid')
    def test_init_no_repo_arg_exception(self):
        self.builder()
        self.assertRaises(MissingArgument, Snapshot, self.ilo)
    def test_init_no_repo_exception(self):
        self.builder()
        self.client.snapshot.get_repository.return_value = {'repo':{'foo':'bar'}}
        self.assertRaises(ActionError, Snapshot, self.ilo, repository='notfound')
    def test_init_no_name_exception(self):
        self.builder()
        self.assertRaises(MissingArgument, Snapshot, self.ilo, repository=testvars.repo_name)
    def test_init_success(self):
        self.builder()
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertEqual(testvars.repo_name, sso.repository)
        self.assertIsNone(sso.state)
    def test_get_state_success(self):
        self.builder()
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        sso.get_state()
        self.assertEqual('SUCCESS', sso.state)
    def test_get_state_fail(self):
        self.builder()
        self.client.snapshot.get.return_value = {'snapshots':[]}
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertRaises(CuratorException, sso.get_state)
    def test_report_state_success(self):
        self.builder()
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        sso.report_state()
        self.assertEqual('SUCCESS', sso.state)
    def test_report_state_other(self):
        self.builder()
        self.client.snapshot.get.return_value = testvars.highly_unlikely
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertRaises(FailedSnapshot, sso.report_state)
    def test_do_dry_run(self):
        self.builder()
        self.client.snapshot.create.return_value = None
        self.client.snapshot.status.return_value = testvars.nosnap_running
        self.client.snapshot.verify_repository.return_value = testvars.verified_nodes
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertIsNone(sso.do_dry_run())
    def test_do_action_success(self):
        self.builder()
        self.client.snapshot.create.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        self.client.snapshot.status.return_value = testvars.nosnap_running
        self.client.snapshot.verify_repository.return_value = testvars.verified_nodes
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertIsNone(sso.do_action())
    def test_do_action_raise_snap_in_progress(self):
        self.builder()
        self.client.snapshot.create.return_value = None
        self.client.snapshot.status.return_value = testvars.snap_running
        self.client.snapshot.verify_repository.return_value = testvars.verified_nodes
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertRaises(SnapshotInProgress, sso.do_action)
    def test_do_action_no_wait_for_completion(self):
        self.builder()
        self.client.snapshot.create.return_value = testvars.generic_task
        self.client.snapshot.status.return_value = testvars.nosnap_running
        self.client.snapshot.verify_repository.return_value = testvars.verified_nodes
        sso = Snapshot(self.ilo, repository=testvars.repo_name,
            name=testvars.snap_name, wait_for_completion=False)
        self.assertIsNone(sso.do_action())
    def test_do_action_raise_on_failure(self):
        self.builder()
        self.client.snapshot.create.return_value = None
        self.client.snapshot.create.side_effect = testvars.fake_fail
        self.client.snapshot.status.return_value = testvars.nosnap_running
        self.client.snapshot.verify_repository.return_value = testvars.verified_nodes
        sso = Snapshot(self.ilo, repository=testvars.repo_name, name=testvars.snap_name)
        self.assertRaises(FailedExecution, sso.do_action)



"""Test count pattern"""

# pylint: disable=C0115, C0116, invalid-name
import os
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')

TIEREDROUTING = {'allocation': {'include': {'_tier_preference': 'data_content'}}}

DELETE_COUNT_PATTERN = (
    '---\n'
    'actions:\n'
    '  1:\n'
    '    description: "Delete indices as filtered"\n'
    '    action: delete_indices\n'
    '    options:\n'
    '      continue_if_exception: False\n'
    '      disable_action: False\n'
    '    filters:\n'
    '      - filtertype: count\n'
    '        pattern: {0}\n'
    '        use_age: {1}\n'
    '        source: {2}\n'
    '        timestring: {3}\n'
    '        reverse: {4}\n'
    '        count: {5}\n'
)


class TestCLICountPattern(CuratorTestCase):
    def test_match_proper_indices(self):
        for i in range(1, 4):
            self.create_index(f'a-{i}')
        for i in range(4, 7):
            self.create_index(f'b-{i}')
        for i in range(5, 9):
            self.create_index(f'c-{i}')
        self.create_index('not_a_match')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            DELETE_COUNT_PATTERN.format(
                r'^(a|b|c)-\d$', 'false', 'name', '\'%Y.%m.%d\'', 'true', 1
            ),
        )
        self.invoke_runner()
        indices = sorted(list(self.client.indices.get(index='*')))
        assert ['a-3', 'b-6', 'c-8', 'not_a_match'] == indices

    def test_match_proper_indices_by_age(self):
        self.create_index('a-2017.10.01')
        self.create_index('a-2017.10.02')
        self.create_index('a-2017.10.03')
        self.create_index('b-2017.09.01')
        self.create_index('b-2017.09.02')
        self.create_index('b-2017.09.03')
        self.create_index('not_a_match')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            DELETE_COUNT_PATTERN.format(
                r'^(a|b)-\d{4}\.\d{2}\.\d{2}$',
                'true',
                'name',
                '\'%Y.%m.%d\'',
                'true',
                1,
            ),
        )
        self.invoke_runner()
        indices = sorted(list(self.client.indices.get(index='*')))
        assert ['a-2017.10.03', 'b-2017.09.03', 'not_a_match'] == indices

    def test_count_indices_by_age_same_age(self):
        key = 'tag'
        value = 'value'
        alloc = 'include'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.allocation_count_test.format(key, value, alloc, False),
        )
        idx1, idx2 = ('c-2017.10.03', 'd-2017.10.03')
        idxlist = [
            'a-2017.10.01',
            'a-2017.10.02',
            'a-2017.10.03',
            'b-2017.10.01',
            'b-2017.10.02',
            'b-2017.10.03',
            'c-2017.10.01',
            'c-2017.10.02',
            'd-2017.10.01',
            'd-2017.10.02',
        ]
        for idx in idxlist:
            self.create_index(idx)
        self.create_index(idx1)
        self.create_index(idx2)
        self.invoke_runner()
        response = self.client.indices.get_settings(index='*')
        for idx in (idx1, idx2):
            assert (
                value
                == response[idx]['settings']['index']['routing']['allocation'][alloc][
                    key
                ]
            )
        for idx in idxlist:
            assert TIEREDROUTING == response[idx]['settings']['index']['routing']



"""Test the Cluster Routing Action"""

# pylint: disable=C0115, C0116, invalid-name
import os
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestCLIClusterRouting(CuratorTestCase):
    def test_allocation_all(self):
        routing_type = 'allocation'
        value = 'all'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.cluster_routing_test.format(routing_type, value),
        )
        self.create_index('my_index')
        self.create_index('not_my_index')
        self.invoke_runner()
        assert testvars.CRA_all == self.client.cluster.get_settings()

    def test_extra_option(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.bad_option_proto_test.format('cluster_routing'),
        )
        self.create_index('my_index')
        self.create_index('not_my_index')
        self.invoke_runner()
        assert 1 == self.result.exit_code



"""Test forcemerge functionality"""

# pylint: disable=C0115, C0116, invalid-name, protected-access
import os
from time import sleep
from curator import IndexList
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestActionFileforceMerge(CuratorTestCase):
    def test_merge(self):
        count = 1
        idx = 'my_index'
        self.create_index(idx)
        self.add_docs(idx)
        ilo1 = IndexList(self.client)
        ilo1.get_segment_counts()
        assert 3 == ilo1.index_info[idx]['segments']
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.forcemerge_test.format(count, 0.9)
        )
        self.invoke_runner()
        ilo2 = IndexList(self.client)
        # This stupid block is only because it can finish testing before
        # the segments have _reported_ as fully merged. This is forcing
        # 3 checks before giving up and reporting the result.
        for _ in range(0, 3):
            self.client.indices.refresh(index=idx)
            ilo2.get_segment_counts()
            if ilo2.index_info[idx]['segments'] == count:
                break
            sleep(1)
        assert count == ilo2.index_info[idx]['segments']

    def test_extra_option(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.bad_option_proto_test.format('forcemerge')
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code


class TestCLIforceMerge(CuratorTestCase):
    COUNT = 1
    IDX = 'my_index'
    ILO = None
    CLI_ARGS = None

    def setup(self):
        self.COUNT = 1
        self.IDX = 'my_index'
        self.create_index(self.IDX)
        self.add_docs(self.IDX)
        self.ILO = IndexList(self.client)
        self.ILO.get_segment_counts()
        assert 3 == self.ILO.index_info[self.IDX]['segments']
        self.CLI_ARGS = self.get_runner_args()
        self.CLI_ARGS += [
            '--config',
            self.args['configfile'],
            'forcemerge',
            '--max_num_segments',
            str(self.COUNT),
            '--delay',
            '0.9',
            '--filter_list',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]

    def wait_for_merge(self):
        # This stupid block is only for the benefit of testing
        # It apparently can finish testing before the segments have _reported_
        # as fully merged.
        # This is forcing 3 checks before giving up and reporting the result.
        for _ in range(0, 3):
            self.client.indices.refresh(index=self.IDX)
            self.ILO.get_segment_counts()
            if self.ILO.index_info[self.IDX]['segments'] == self.COUNT:
                break
            sleep(1)

    def test_merge(self):
        self.setup()
        assert 0 == self.run_subprocess(
            self.CLI_ARGS, logname='TestCLIforceMerge.test_merge'
        )
        self.wait_for_merge()
        assert self.COUNT == self.ILO.index_info[self.IDX]['segments']
        self.ILO = None
        self.CLI_ARGS = None

    def test_empty_list1(self):
        """Test with an empty list and ignore_empty_list unset"""
        self.setup()
        self.client.indices.forcemerge(index=self.IDX, max_num_segments=1)
        self.wait_for_merge()
        assert 1 == self.run_subprocess(
            self.CLI_ARGS, logname='TestCLIforceMerge.test_empty_list1'
        )
        self.ILO = None
        self.CLI_ARGS = None

    def test_empty_list2(self):
        """Test with an empty list and ignore_empty_list = True"""
        self.setup()
        self.client.indices.forcemerge(index=self.IDX, max_num_segments=1)
        self.wait_for_merge()
        self.CLI_ARGS = self.get_runner_args()
        self.CLI_ARGS += [
            '--config',
            self.args['configfile'],
            'forcemerge',
            '--max_num_segments',
            str(self.COUNT),
            '--delay',
            '0.9',
            '--ignore_empty_list',
            '--filter_list',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]
        assert 0 == self.run_subprocess(
            self.CLI_ARGS, logname='TestCLIforceMerge.test_empty_list2'
        )
        self.ILO = None
        self.CLI_ARGS = None



"""Delete Index and Delete Snapshot Singletons"""

# pylint: disable=R0913,R0917
import click
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import validate_filter_json


# Indices
@click.command()
@click.option(
    '--search_pattern',
    type=str,
    default='*',
    help='Elasticsearch Index Search Pattern',
)
@click.option(
    '--ignore_empty_list',
    is_flag=True,
    help='Do not raise exception if there are no actionable indices',
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True,
)
@click.option(
    '--filter_list',
    callback=validate_filter_json,
    help='JSON array of filters selecting indices to act on.',
    required=True,
)
@click.pass_context
def delete_indices(
    ctx, search_pattern, ignore_empty_list, allow_ilm_indices, filter_list
):
    """
    Delete Indices
    """
    # ctx.info_name is the name of the function or name specified in @click.command
    # decorator
    action = CLIAction(
        'delete_indices',
        ctx.obj['configdict'],
        {'search_pattern': search_pattern, 'allow_ilm_indices': allow_ilm_indices},
        filter_list,
        ignore_empty_list,
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])


# Snapshots
@click.command()
@click.option('--repository', type=str, required=True, help='Snapshot repository name')
@click.option('--retry_count', type=int, help='Number of times to retry (max 3)')
@click.option('--retry_interval', type=int, help='Time in seconds between retries')
@click.option(
    '--ignore_empty_list',
    is_flag=True,
    help='Do not raise exception if there are no actionable snapshots',
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True,
)
@click.option(
    '--include_hidden/--no-include_hidden',
    help='Allow Curator to operate on hidden indices (and data_streams).',
    default=False,
    show_default=True,
)
@click.option(
    '--filter_list',
    callback=validate_filter_json,
    help='JSON array of filters selecting snapshots to act on.',
    required=True,
)
@click.pass_context
def delete_snapshots(
    ctx,
    repository,
    retry_count,
    retry_interval,
    ignore_empty_list,
    allow_ilm_indices,
    include_hidden,
    filter_list,
):
    """
    Delete Snapshots
    """
    manual_options = {
        'retry_count': retry_count,
        'retry_interval': retry_interval,
        'allow_ilm_indices': allow_ilm_indices,
        'include_hidden': include_hidden,
    }
    # ctx.info_name is the name of the function or name specified in @click.command
    # decorator
    action = CLIAction(
        'delete_snapshots',
        ctx.obj['configdict'],
        manual_options,
        filter_list,
        ignore_empty_list,
        repository=repository,
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""test_action_reindex"""
# pylint: disable=missing-function-docstring, missing-class-docstring, protected-access, attribute-defined-outside-init
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import Reindex
from curator.exceptions import ConfigurationError, CuratorException, FailedExecution, NoIndices
from curator import IndexList
# Get test variables and constants from a single source
from . import testvars

class TestActionReindex(TestCase):
    VERSION = {'version': {'number': '8.0.0'} }
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = self.VERSION
        self.client.cat.indices.return_value = testvars.state_four
        self.client.indices.get_settings.return_value = testvars.settings_four
        self.client.indices.stats.return_value = testvars.stats_four
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)
    def test_init_bad_ilo(self):
        self.assertRaises(TypeError, Reindex, 'foo', 'invalid')
    def test_init_raise_bad_request_body(self):
        self.builder()
        self.assertRaises(ConfigurationError, Reindex, self.ilo, 'invalid')
    def test_init_raise_local_migration_no_prefix_or_suffix(self):
        self.builder()
        self.assertRaises(ConfigurationError, Reindex, self.ilo, testvars.reindex_migration)
    def test_init(self):
        self.builder()
        rio = Reindex(self.ilo, testvars.reindex_basic)
        self.assertEqual(self.ilo, rio.index_list)
        self.assertEqual(self.client, rio.client)
    def test_do_dry_run(self):
        self.builder()
        rio = Reindex(self.ilo, testvars.reindex_basic)
        self.assertIsNone(rio.do_dry_run())
    def test_replace_index_list(self):
        self.builder()
        rio = Reindex(self.ilo, testvars.reindex_replace)
        self.assertEqual(rio.index_list.indices, rio.body['source']['index'])
    def test_reindex_with_wait(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        rio = Reindex(self.ilo, testvars.reindex_basic)
        self.assertIsNone(rio.do_action())
    def test_reindex_with_wait_zero_total(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task_zero_total
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        rio = Reindex(self.ilo, testvars.reindex_basic)
        self.assertIsNone(rio.do_action())
    def test_reindex_with_wait_zero_total_fail(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.side_effect = testvars.fake_fail
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        rio = Reindex(self.ilo, testvars.reindex_basic)
        self.assertRaises(CuratorException, rio.do_action)
    def test_reindex_without_wait(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        rio = Reindex(self.ilo, testvars.reindex_basic,
            wait_for_completion=False)
        self.assertIsNone(rio.do_action())
    def test_reindex_timedout(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.incomplete_task
        rio = Reindex(self.ilo, testvars.reindex_basic,
             max_wait=1, wait_interval=1)
        self.assertRaises(FailedExecution, rio.do_action)
    def test_remote_with_no_host_key(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        badval = {
            'source': {
                'index': 'irrelevant',
                'remote': {'wrong': 'invalid'}
            },
            'dest': { 'index': 'other_index' }
        }
        self.assertRaises(
            ConfigurationError, Reindex, self.ilo, badval)
    def test_remote_with_bad_host(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        badval = {
            'source': {
                'index': 'irrelevant',
                'remote': {'host': 'invalid'}
            },
            'dest': { 'index': 'other_index' }
        }
        self.assertRaises(
            ConfigurationError, Reindex, self.ilo, badval)
    def test_remote_with_bad_url(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        badval = {
            'source': {
                'index': 'irrelevant',
                'remote': {'host': 'asdf://hostname:1234'}
            },
            'dest': { 'index': 'other_index' }
        }
        self.assertRaises(
            ConfigurationError, Reindex, self.ilo, badval)
    def test_remote_with_bad_connection(self):
        self.builder()
        self.client.reindex.return_value = testvars.generic_task
        self.client.tasks.get.return_value = testvars.completed_task
        # After building ilo, we need a different return value
        self.client.indices.get_settings.return_value = {'other_index':{}}
        badval = {
            'source': {
                'index': 'REINDEX_SELECTION',
                'remote': {'host': 'https://example.org:XXXX'}
            },
            'dest': { 'index': 'other_index' }
        }
        urllib3 = Mock()
        urllib3.util.retry.side_effect = testvars.fake_fail
        self.assertRaises(Exception, Reindex, self.ilo, badval)
    def test_init_raise_empty_source_list(self):
        self.builder()
        badval = {
            'source': { 'index': [] },
            'dest': { 'index': 'other_index' }
        }
        rio = Reindex(self.ilo, badval)
        self.assertRaises(NoIndices, rio.do_action)



"""test_action_open"""
# pylint: disable=missing-function-docstring, missing-class-docstring, protected-access, attribute-defined-outside-init
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import Open
from curator.exceptions import FailedExecution
from curator import IndexList
# Get test variables and constants from a single source
from . import testvars

class TestActionOpen(TestCase):
    VERSION = {'version': {'number': '8.0.0'} }
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = self.VERSION
        self.client.cat.indices.return_value = testvars.state_four
        self.client.indices.get_settings.return_value = testvars.settings_four
        self.client.indices.stats.return_value = testvars.stats_four
        self.client.indices.exists_alias.return_value = False
        self.client.indices.open.return_value = None
        self.ilo = IndexList(self.client)
    def test_init_raise(self):
        self.assertRaises(TypeError, Open, 'invalid')
    def test_init(self):
        self.builder()
        opn = Open(self.ilo)
        self.assertEqual(self.ilo, opn.index_list)
        self.assertEqual(self.client, opn.client)
    def test_do_dry_run(self):
        self.builder()
        self.ilo.filter_opened()
        opn = Open(self.ilo)
        self.assertEqual(['c-2016.03.05'], opn.index_list.indices)
        self.assertIsNone(opn.do_dry_run())
    def test_do_action(self):
        self.builder()
        self.ilo.filter_opened()
        opn = Open(self.ilo)
        self.assertEqual(['c-2016.03.05'], opn.index_list.indices)
        self.assertIsNone(opn.do_action())
    def test_do_action_raises_exception(self):
        self.builder()
        self.client.indices.open.side_effect = testvars.fake_fail
        opn = Open(self.ilo)
        self.assertRaises(FailedExecution, opn.do_action)



"""Integration tests of the Close action class"""

# pylint: disable=C0115, C0116, invalid-name
import os
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')

MET = 'metadata'
IDC = 'indices'
STA = 'state'


def get_state(client, data):
    """Return cluster state data"""
    return client.cluster.state(index=data, metric=MET)


class TestActionFileClose(CuratorTestCase):
    def test_close_opened(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.optionless_proto.format('close')
        )
        idx1, idx2 = ('dummy', 'my_index')
        indices = f'{idx1},{idx2}'
        self.create_index(idx1)
        self.create_index(idx2)
        self.invoke_runner()
        state = get_state(self.client, indices)
        assert 'close' != state[MET][IDC][idx1][STA]
        assert 'close' == state[MET][IDC][idx2][STA]

    def test_close_closed(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.optionless_proto.format('close')
        )
        idx1, idx2 = ('dummy', 'my_index')
        indices = f'{idx1},{idx2}'
        self.create_index(idx2)
        # pylint: disable=E1123
        self.client.indices.close(
            index=idx2, ignore_unavailable=True, wait_for_active_shards=0
        )
        self.create_index(idx1)
        self.invoke_runner()
        state = get_state(self.client, indices)
        assert 'close' != state[MET][IDC][idx1][STA]
        assert 'close' == state[MET][IDC][idx2][STA]

    def test_close_delete_aliases(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.close_delete_aliases)
        # Create aliases first
        alias = 'testalias'
        idxtuple = ('dummy', 'my_index', 'my_other')
        idx1, idx2, idx3 = idxtuple
        indices = None
        for idx in idxtuple:
            self.create_index(idx)
            indices = f'{indices},{idx}' if indices else f'{idx}'
        self.client.indices.put_alias(index=f'{idx1},{idx2}', name=alias)
        precheck = {idx1: {"aliases": {alias: {}}}, idx2: {"aliases": {alias: {}}}}
        assert precheck == self.client.indices.get_alias(name=alias)
        # Now close `index` with delete_aliases=True (dummy stays open)
        self.invoke_runner()
        state = get_state(self.client, indices)
        for idx in (idx2, idx3):
            assert 'close' == state[MET][IDC][idx][STA]
        # Now open the indices and verify that the alias is still gone.
        self.client.indices.open(index=idx2)
        assert {idx1: {"aliases": {alias: {}}}} == self.client.indices.get_alias(
            name=alias
        )

    def test_close_skip_flush(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.close_skip_flush)
        idx1, idx2 = ('dummy', 'my_index')
        indices = f'{idx1},{idx2}'
        self.create_index(idx1)
        # Disable shard allocation to make my_index go red
        disable_allocation = {"cluster.routing.allocation.enable": "none"}
        self.client.cluster.put_settings(transient=disable_allocation)
        self.create_index(idx2, wait_for_yellow=False, wait_for_active_shards=0)
        self.invoke_runner()
        state = get_state(self.client, indices)
        assert 'close' != state[MET][IDC][idx1][STA]
        assert 'close' == state[MET][IDC][idx2][STA]

    def test_extra_option(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.bad_option_proto_test.format('close')
        )
        idx1, idx2 = ('dummy', 'my_index')
        indices = f'{idx1},{idx2}'
        self.create_index(idx1)
        self.create_index(idx2)
        self.invoke_runner()
        state = get_state(self.client, indices)
        for idx in (idx1, idx2):
            assert 'close' != state[MET][IDC][idx][STA]
        assert 1 == self.result.exit_code


class TestCLIClose(CuratorTestCase):
    def test_close_delete_aliases(self):
        # Create aliases first
        alias = 'testalias'
        idx1, idx2, idx3 = ('my_index', 'dummy', 'my_other')
        indices = f'{idx1},{idx2},{idx3}'
        self.create_index(idx1)
        self.create_index(idx2)
        self.create_index(idx3)
        self.client.indices.put_alias(index=f'{idx1},{idx2}', name=alias)
        precheck = {idx1: {"aliases": {alias: {}}}, idx2: {"aliases": {alias: {}}}}
        assert precheck == self.client.indices.get_alias(name=alias)
        # Now close `index` with delete_aliases=True idx2 stays open)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'close',
            '--delete_aliases',
            '--filter_list',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]
        assert 0 == self.run_subprocess(
            args, logname='TestCLIClose.test_close_delete_aliases'
        )
        state = get_state(self.client, indices)
        for idx in (idx1, idx3):
            assert 'close' == state[MET][IDC][idx][STA]
        # Now open the indices and verify that the alias is still gone.
        self.client.indices.open(index=idx1)
        assert {idx2: {"aliases": {alias: {}}}} == self.client.indices.get_alias(
            name=alias
        )

    def test_close_skip_flush(self):
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'close',
            '--skip_flush',
            '--filter_list',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]
        idx1, idx2 = ('my_index', 'dummy')
        indices = f'{idx1},{idx2}'
        self.create_index(idx1)
        self.create_index(idx2)
        assert 0 == self.run_subprocess(
            args, logname='TestCLIClose.test_close_skip_flush'
        )
        state = get_state(self.client, indices)
        assert 'close' == state[MET][IDC][idx1][STA]



"""Test the Create Index action"""

# pylint: disable=C0115, C0116, invalid-name
import os
from curator import IndexList
from curator.helpers.date_ops import parse_date_pattern
from curator.helpers.getters import get_indices
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestCLICreateIndex(CuratorTestCase):
    def test_plain(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.create_index.format('testing')
        )
        assert not get_indices(self.client)
        self.invoke_runner()
        self.assertEqual(['testing'], get_indices(self.client))
        assert ['testing'] == get_indices(self.client)

    def test_with_extra_settings(self):
        idx = 'testing'
        alias = 'aliasname'
        mapkey1 = 'meep'
        mapval1 = 'integer'
        mapkey2 = 'beep'
        mapval2 = 'keyword'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.create_index_with_extra_settings.format(
                idx, alias, mapkey1, mapval1, mapkey2, mapval2
            ),
        )
        assert not get_indices(self.client)
        self.invoke_runner()
        ilo = IndexList(self.client)
        ilo.get_index_settings()
        aliases = self.client.indices.get_alias(name=alias)
        mapping = self.client.indices.get_mapping(index=idx)
        assert [idx] == ilo.indices
        assert '1' == ilo.index_info[idx]['number_of_shards']
        assert '0' == ilo.index_info[idx]['number_of_replicas']
        assert mapping[idx]['mappings']['properties'][mapkey1] == {'type': mapval1}
        assert mapping[idx]['mappings']['properties'][mapkey2] == {'type': mapval2}
        assert aliases[idx]['aliases'] == {alias: {'is_write_index': True}}

    def test_with_strftime(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.create_index.format('testing-%Y.%m.%d')
        )
        assert not get_indices(self.client)
        idx = parse_date_pattern('testing-%Y.%m.%d')
        self.invoke_runner()
        assert [idx] == get_indices(self.client)

    def test_with_date_math(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.create_index.format('<testing-{now/d}>')
        )
        assert not get_indices(self.client)
        idx = parse_date_pattern('testing-%Y.%m.%d')
        self.invoke_runner()
        assert [idx] == get_indices(self.client)

    def test_extra_option(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.bad_option_proto_test.format('create_index'),
        )
        self.invoke_runner()
        assert not get_indices(self.client)
        assert 1 == self.result.exit_code

    def test_already_existing_fail(self):
        idx = 'testing'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.create_index.format(idx))
        self.create_index(idx)
        self.invoke_runner()
        assert [idx] == get_indices(self.client)
        assert 1 == self.result.exit_code

    def test_already_existing_pass(self):
        config = (
            '---\n'
            'actions:\n'
            '  1:\n'
            '    description: "Create index as named"\n'
            '    action: create_index\n'
            '    options:\n'
            '      name: {0}\n'
            '      ignore_existing: true\n'
        )
        idx = 'testing'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], config.format(idx))
        self.create_index(idx)
        self.invoke_runner()
        assert [idx] == get_indices(self.client)
        assert 0 == self.result.exit_code

    def test_incorrect_mapping_fail_with_propper_error(self):
        config = (
            '---\n'
            'actions:\n'
            '  1:\n'
            '    description: "Create index as named"\n'
            '    action: create_index\n'
            '    options:\n'
            '      name: {0}\n'
            '      extra_settings:\n'
            '        mappings:\n'
            '          properties:\n'
            '            name: ["test"]\n'
        )
        idx = 'testing'
        self.write_config(
            self.args['configfile'], testvars.none_logging_config.format(HOST)
        )
        self.write_config(self.args['actionfile'], config.format(idx))
        self.invoke_runner()
        assert not get_indices(self.client)
        assert 'mapper_parsing_exception' in self.result.stderr_bytes.decode('utf-8')
        assert 1 == self.result.exit_code



"""Test Alias action"""

# pylint: disable=C0115, C0116, invalid-name
import os
import logging
from datetime import datetime, timedelta, timezone
import pytest
from elasticsearch9.exceptions import NotFoundError
from . import CuratorTestCase
from . import testvars

LOGGER = logging.getLogger(__name__)

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestActionFileAlias(CuratorTestCase):
    def test_add_only(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_add_only.format(alias)
        )
        self.create_index('my_index')
        self.create_index('dummy')
        self.client.indices.put_alias(index='dummy', name=alias)
        self.invoke_runner()
        assert 2 == len(self.client.indices.get_alias(name=alias))

    def test_add_only_with_extra_settings(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.alias_add_only_with_extra_settings.format(alias),
        )
        self.create_index('my_index')
        self.invoke_runner()
        expected = {
            'my_index': {
                'aliases': {'testalias': {'filter': {'term': {'user': 'kimchy'}}}}
            }
        }
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_alias_remove_only(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_remove_only.format(alias)
        )
        idx1, idx2 = ('my_index', 'dummy')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx2, name=alias)
        self.invoke_runner()
        expected = {idx2: {'aliases': {}}, idx1: {'aliases': {}}}
        result = self.client.indices.get_alias(index=f'{idx1},{idx2}')
        assert expected == result.body

    def test_add_only_skip_closed(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_add_only.format(alias)
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx2)
        self.client.indices.close(index=idx2, wait_for_active_shards=0)
        self.create_index(idx1)
        self.client.indices.put_alias(index=idx1, name=alias)
        self.invoke_runner()
        assert 2 == len(self.client.indices.get_alias(name=alias))

    def test_add_and_remove(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_add_remove.format(alias)
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias)
        self.invoke_runner()
        expected = {idx2: {'aliases': {alias: {}}}}
        success = False
        result = self.client.indices.get_alias(name=alias)
        tries = 0
        max_tries = 10
        while not success and tries < max_tries:
            if expected == result.body:
                success = True
            else:
                tries += 1
                result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_add_and_remove_datemath(self):
        alias = '<testalias-{now-1d/d}>'
        _ = (datetime.now(timezone.utc) - timedelta(days=1)).strftime('%Y.%m.%d')
        alias_parsed = f"testalias-{_}"
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_add_remove.format(alias)
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias_parsed)
        self.invoke_runner()
        expected = {idx2: {'aliases': {alias_parsed: {}}}}
        result = self.client.indices.get_alias(name=alias_parsed)
        assert expected == result.body

    def test_add_with_empty_remove(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_add_with_empty_remove.format(alias)
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias)
        self.invoke_runner()
        expected = {idx1: {'aliases': {alias: {}}}, idx2: {'aliases': {alias: {}}}}
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_remove_with_empty_add(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_remove_with_empty_add.format(alias)
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=f'{idx1},{idx2}', name=alias)
        self.invoke_runner()
        expected = {idx2: {'aliases': {alias: {}}}}
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_add_with_empty_list(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.alias_add_remove_empty.format(alias, 'du', 'rickroll'),
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias)
        self.invoke_runner()
        expected = {idx1: {'aliases': {alias: {}}}}
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_remove_with_empty_list(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.alias_add_remove_empty.format(alias, 'rickroll', 'my'),
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias)
        self.invoke_runner()
        expected = {idx1: {'aliases': {alias: {}}}}
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_remove_index_not_in_alias(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.alias_remove_index_not_there.format(alias, 'my'),
        )
        idx1, idx2 = ('my_index1', 'my_index2')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias)
        self.invoke_runner()
        with pytest.raises(NotFoundError):
            self.client.indices.get_alias(name=alias)
        assert 0 == self.result.exit_code

    def test_no_add_remove(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.alias_no_add_remove.format(alias)
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_no_alias(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.alias_no_alias)
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_extra_options(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.bad_option_proto_test.format('alias')
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_add_and_remove_sorted(self):
        alias = 'testalias'
        alias_add_remove = (
            '---\n'
            'actions:\n'
            '  1:\n'
            '    description: "Add/remove specified indices from designated alias"\n'
            '    action: alias\n'
            '    options:\n'
            '      name: {0}\n'
            '      continue_if_exception: False\n'
            '      disable_action: False\n'
            '    add:\n'
            '      filters:\n'
            '        - filtertype: pattern\n'
            '          kind: prefix\n'
            '          value: dum\n'
            '    remove:\n'
            '      filters:\n'
            '        - filtertype: pattern\n'
            '          kind: prefix\n'
            '          value: my\n'
        )
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], alias_add_remove.format(alias))
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx2, name=alias)
        self.invoke_runner()
        expected = {idx1: {'aliases': {alias: {}}}}
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body


class TestCLIAlias(CuratorTestCase):
    """Test CLI Singleton Alias action"""

    def test_add_and_remove_alias(self):
        """test_add_and_remove_alias"""
        alias = 'testalias'
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx2, name=alias)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'alias',
            '--name',
            alias,
            '--add',
            '{"filtertype":"pattern","kind":"prefix","value":"dum"}',
            '--remove',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]
        assert 0 == self.run_subprocess(args)
        expected = {idx1: {'aliases': {alias: {}}}}
        success = False
        result = self.client.indices.get_alias(name=alias)
        tries = 0
        max_tries = 1
        while not success and tries < max_tries:
            if expected == result.body:
                success = True
            else:
                tries += 1
                result = self.client.indices.get_alias(name=alias)
        assert success

    def test_warn_if_no_indices(self):
        """test_warn_if_no_indices"""
        alias = 'testalias'
        idx1, idx2 = ('dummy1', 'dummy2')
        self.create_index(idx1)
        self.create_index(idx2)
        self.client.indices.put_alias(index=idx1, name=alias)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'alias',
            '--name',
            alias,
            '--add',
            '{"filtertype":"none"}',
            '--remove',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
            '--warn_if_no_indices',
        ]
        LOGGER.debug('ARGS = %s', args)
        assert 0 == self.run_subprocess(args)
        expected = {idx1: {'aliases': {alias: {}}}, idx2: {'aliases': {alias: {}}}}
        result = self.client.indices.get_alias(name=alias)
        assert expected == result.body

    def test_exit_1_on_empty_list(self):
        """test_exit_1_on_empty_list"""
        alias = 'testalias'
        self.create_index('dummy')
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'alias',
            '--name',
            alias,
            '--add',
            '{"filtertype":"pattern","kind":"prefix","value":"dum","exclude":false}',
            '--remove',
            '{"filtertype":"pattern","kind":"prefix","value":"my","exclude":false}',
        ]
        assert 1 == self.run_subprocess(
            args, logname='TestCLIAlias.test_exit_1_on_empty_list'
        )



"""Test snapshot restore functionality"""

# pylint: disable=missing-function-docstring, missing-class-docstring, line-too-long
import os
import time
from curator.helpers.getters import get_indices, get_snapshot
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')

# '      repository: {0}\n'
# '      name: {1}\n'
# '      indices: {2}\n'
# '      include_aliases: {3}\n'
# '      ignore_unavailable: {4}\n'
# '      include_global_state: {5}\n'
# '      partial: {6}\n'
# '      rename_pattern: {7}\n'
# '      rename_replacement: {8}\n'
# '      extra_settings: {9}\n'
# '      wait_for_completion: {10}\n'
# '      skip_repo_fs_check: {11}\n'
# '      timeout_override: {12}\n'
# '      wait_interval: {13}\n'
# '      max_wait: {14}\n'


class TestActionFileRestore(CuratorTestCase):
    """Test file-based configuration restore operations"""

    def test_restore(self):
        """Test restore action"""
        indices = []
        for i in range(1, 4):
            self.add_docs(f'my_index{i}')
            indices.append(f'my_index{i}')
        snap_name = 'snapshot1'
        self.create_snapshot(snap_name, ','.join(indices))
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        self.assertEqual(1, len(snapshot['snapshots']))
        assert 1 == len(snapshot['snapshots'])
        self.client.indices.delete(index=','.join(indices))
        self.assertEqual([], get_indices(self.client))
        assert not get_indices(self.client)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.restore_snapshot_proto.format(
                self.args['repository'],
                snap_name,
                indices,
                False,
                False,
                True,
                False,
                ' ',
                ' ',
                ' ',
                True,
                False,
                301,
                1,
                3,
            ),
        )
        self.invoke_runner()
        restored_indices = sorted(get_indices(self.client))
        self.assertEqual(indices, restored_indices)
        assert indices == restored_indices
        # The test runs so fast that it tries to execute the cleanup step
        # and delete the repository before Elasticsearch is actually ready
        time.sleep(0.5)

    def test_restore_with_rename(self):
        """Test restore action with renaming enabled"""
        indices = []
        for i in range(1, 4):
            self.add_docs(f'my_index{i}')
            indices.append(f'my_index{i}')
        snap_name = 'snapshot1'
        self.create_snapshot(snap_name, ','.join(indices))
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        time.sleep(1)
        self.assertEqual(1, len(snapshot['snapshots']))
        assert 1 == len(snapshot['snapshots'])
        self.client.indices.delete(index=','.join(indices))
        self.assertEqual([], get_indices(self.client))
        assert not get_indices(self.client)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.restore_snapshot_proto.format(
                self.args['repository'],
                snap_name,
                indices,
                False,
                False,
                True,
                False,
                'my_index(.+)',
                'new_index$1',
                ' ',
                True,
                False,
                301,
                1,
                3,
            ),
        )
        self.invoke_runner()
        time.sleep(1)
        restored_indices = sorted(get_indices(self.client))
        self.assertEqual(['new_index1', 'new_index2', 'new_index3'], restored_indices)
        assert ['new_index1', 'new_index2', 'new_index3'] == restored_indices
        # The test runs so fast that it tries to execute the cleanup step
        # and delete the repository before Elasticsearch is actually ready
        time.sleep(1)

    def test_restore_wildcard(self):
        """Test restore with wildcard"""
        indices = []
        my_indices = []
        wildcard = ['my_*']
        for i in range(1, 4):
            for prefix in ['my_', 'not_my_']:
                self.add_docs(f'{prefix}index{i}')
                indices.append(f'{prefix}index{i}')
                if prefix == 'my_':
                    my_indices.append(f'{prefix}index{i}')
        snap_name = 'snapshot1'
        self.create_snapshot(snap_name, ','.join(indices))
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        self.assertEqual(1, len(snapshot['snapshots']))
        assert 1 == len(snapshot['snapshots'])
        self.client.indices.delete(index=','.join(indices))
        self.assertEqual([], get_indices(self.client))
        assert not get_indices(self.client)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.restore_snapshot_proto.format(
                self.args['repository'],
                snap_name,
                wildcard,
                False,
                False,
                True,
                False,
                ' ',
                ' ',
                ' ',
                True,
                False,
                301,
                1,
                3,
            ),
        )
        self.invoke_runner()
        restored_indices = sorted(get_indices(self.client))
        self.assertEqual(my_indices, restored_indices)
        assert my_indices == restored_indices
        # The test runs so fast that it tries to execute the cleanup step
        # and delete the repository before Elasticsearch is actually ready
        time.sleep(0.5)


class TestCLIRestore(CuratorTestCase):
    def test_restore(self):
        indices = []
        for i in range(1, 4):
            self.add_docs(f'my_index{i}')
            indices.append(f'my_index{i}')
        snap_name = 'snapshot1'
        self.create_snapshot(snap_name, ','.join(indices))
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        self.assertEqual(1, len(snapshot['snapshots']))
        assert 1 == len(snapshot['snapshots'])
        self.client.indices.delete(index=f'{",".join(indices)}')
        self.assertEqual([], get_indices(self.client))
        assert not get_indices(self.client)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'restore',
            '--repository',
            self.args['repository'],
            '--name',
            snap_name,
            '--index',
            indices[0],
            '--index',
            indices[1],
            '--index',
            indices[2],
            '--wait_interval',
            '1',
            '--max_wait',
            '3',
            '--filter_list',
            '{"filtertype":"none"}',
        ]
        # self.assertEqual(
        #     0, self.run_subprocess(args, logname='TestCLIRestore.test_restore')
        # )
        assert 0 == self.run_subprocess(args, logname='TestCLIRestore.test_restore')
        restored_indices = sorted(get_indices(self.client))
        self.assertEqual(indices, restored_indices)
        assert indices == restored_indices
        # The test runs so fast that it tries to execute the cleanup step
        # and delete the repository before Elasticsearch is actually ready
        time.sleep(0.5)



"""test_action_shrink"""

# pylint: disable=C0103,C0115,C0116,W0201,W0212
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import Shrink
from curator.exceptions import ActionError, ConfigurationError
from curator import IndexList

# Get test variables and constants from a single source
from . import testvars


class TestActionShrink_ilo(TestCase):
    def test_init_raise_bad_client(self):
        self.assertRaises(TypeError, Shrink, 'invalid')


class TestActionShrink_extra_settings(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)

    def test_extra_settings_1(self):
        self.builder()
        self.assertRaises(
            ConfigurationError,
            Shrink,
            self.ilo,
            extra_settings={'settings': {'foobar'}},
        )

    def test_extra_settings_2(self):
        self.builder()
        self.assertRaises(
            ConfigurationError, Shrink, self.ilo, extra_settings={'foobar'}
        )


class TestActionShrink_data_node(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.node_name = 'node_name'
        self.node_id = 'my_node'
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'name': self.node_name}}
        }
        self.ilo = IndexList(self.client)
        self.shrink = Shrink(self.ilo)

    def test_non_data_node(self):
        self.builder()
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['ingest'], 'name': self.node_name}}
        }
        self.assertFalse(self.shrink._data_node(self.node_id))

    def test_data_node(self):
        self.builder()
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['data'], 'name': self.node_name}}
        }
        self.assertTrue(self.shrink._data_node(self.node_id))


class TestActionShrink_exclude_node(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.node_name = 'node_name'
        self.ilo = IndexList(self.client)

    def test_positive(self):
        self.builder()
        node_filters = {'exclude_nodes': [self.node_name]}
        shrink = Shrink(self.ilo, node_filters=node_filters)
        self.assertTrue(shrink._exclude_node(self.node_name))

    def test_negative(self):
        self.builder()
        node_filters = {'exclude_nodes': ['not_this_node']}
        shrink = Shrink(self.ilo, node_filters=node_filters)
        self.assertFalse(shrink._exclude_node(self.node_name))


class TestActionShrink_qualify_single_node(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.node_name = 'node_name'
        self.node_id = 'my_node'
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'name': self.node_name}}
        }
        self.ilo = IndexList(self.client)

    def test_positive(self):
        self.builder()
        byte_count = 123456
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['data'], 'name': self.node_name}}
        }
        self.client.nodes.stats.return_value = {
            'nodes': {
                self.node_id: {
                    'fs': {
                        'data': ['one'],
                        'total': {'available_in_bytes': byte_count},
                    },
                    'name': self.node_name,
                }
            }
        }
        shrink = Shrink(self.ilo, shrink_node=self.node_name)
        shrink.qualify_single_node()
        self.assertEqual(byte_count, shrink.shrink_node_avail)

    def test_not_found(self):
        self.builder()
        shrink = Shrink(self.ilo, shrink_node='not_me')
        self.assertRaises(ConfigurationError, shrink.qualify_single_node)

    def test_excluded(self):
        self.builder()
        node_filters = {'exclude_nodes': [self.node_name]}
        shrink = Shrink(self.ilo, shrink_node=self.node_name, node_filters=node_filters)
        self.assertRaises(ConfigurationError, shrink.qualify_single_node)

    def test_non_data_node(self):
        self.builder()
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['ingest'], 'name': self.node_name}}
        }
        shrink = Shrink(self.ilo, shrink_node=self.node_name)
        self.assertRaises(ActionError, shrink.qualify_single_node)


class TestActionShrink_most_available_node(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.node_name = 'node_name'
        self.node_id = 'my_node'
        self.byte_count = 123456
        self.client.nodes.stats.return_value = {
            'nodes': {self.node_id: {'name': self.node_name}}
        }
        self.ilo = IndexList(self.client)

    def test_excluded(self):
        self.builder()
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['data']}}
        }
        self.client.nodes.stats.return_value = {
            'nodes': {
                self.node_id: {
                    'fs': {
                        'data': ['one'],
                        'total': {'available_in_bytes': self.byte_count},
                    },
                    'name': self.node_name,
                }
            }
        }
        node_filters = {'exclude_nodes': [self.node_name]}
        shrink = Shrink(self.ilo, node_filters=node_filters)
        shrink.most_available_node()
        self.assertIsNone(shrink.shrink_node_name)


class TestActionShrink_route_index(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)
        self.shrink = Shrink(self.ilo)

    def test_raises(self):
        self.builder()
        self.client.indices.put_settings.side_effect = testvars.fake_fail
        self.assertRaises(
            Exception,
            self.shrink.route_index,
            'index',
            'exclude',
            '_name',
            'not_my_node',
        )


class TestActionShrink_dry_run(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.node_name = 'node_name'
        self.node_id = 'my_node'
        self.byte_count = 123456
        self.client.nodes.stats.return_value = {
            'nodes': {
                self.node_id: {
                    'fs': {
                        'data': ['one'],
                        'total': {'available_in_bytes': self.byte_count},
                    },
                    'name': self.node_name,
                }
            }
        }
        self.ilo = IndexList(self.client)

    def test_dry_run(self):
        self.builder()
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['data'], 'name': self.node_name}}
        }
        self.client.indices.get.return_value = {
            testvars.named_index: {'settings': {'index': {'number_of_shards': 2}}}
        }
        shrink = Shrink(
            self.ilo,
            shrink_node=self.node_name,
            post_allocation={
                'allocation_type': 'require',
                'key': '_name',
                'value': self.node_name,
            },
        )
        self.assertIsNone(shrink.do_dry_run())

    def test_dry_run_raises(self):
        self.builder()
        self.client.nodes.info.return_value = {
            'nodes': {self.node_id: {'roles': ['data'], 'name': self.node_name}}
        }
        self.client.indices.get.side_effect = testvars.fake_fail
        shrink = Shrink(self.ilo, shrink_node=self.node_name)
        self.assertRaises(Exception, shrink.do_dry_run)


class TestActionShrink_various(TestCase):
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = {'version': {'number': '8.0.0'}}
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.node_name = 'node_name'
        self.node_id = 'my_node'
        self.byte_count = 1239132959
        self.client.nodes.stats.return_value = {
            'nodes': {
                self.node_id: {
                    'fs': {
                        'data': ['one'],
                        'total': {'available_in_bytes': self.byte_count},
                    },
                    'name': self.node_name,
                }
            }
        }
        self.ilo = IndexList(self.client)

    def test_target_exists(self):
        self.builder()
        self.client.indices.exists.return_value = True
        shrink = Shrink(self.ilo)
        self.assertRaises(
            ActionError, shrink._check_target_exists, testvars.named_index
        )

    def test_doc_count(self):
        self.builder()
        too_many = 2147483520
        self.client.indices.stats.return_value = {
            'indices': {
                testvars.named_index: {'primaries': {'docs': {'count': too_many}}}
            }
        }
        shrink = Shrink(self.ilo)
        self.assertRaises(ActionError, shrink._check_doc_count, testvars.named_index)

    def test_shard_count(self):
        self.builder()
        src_shards = 2
        shrink = Shrink(self.ilo, number_of_shards=src_shards)
        self.assertRaises(
            ActionError, shrink._check_shard_count, testvars.named_index, src_shards
        )

    def test_shard_factor(self):
        self.builder()
        src_shards = 5
        shrink = Shrink(self.ilo, number_of_shards=3)
        self.assertRaises(
            ActionError, shrink._check_shard_factor, testvars.named_index, src_shards
        )

    def test_check_all_shards(self):
        self.builder()
        self.client.cluster.state.return_value = {
            'routing_table': {
                'indices': {
                    testvars.named_index: {
                        'shards': {
                            '0': [
                                {
                                    'index': testvars.named_index,
                                    'node': 'not_this_node',
                                    'primary': True,
                                    'shard': 0,
                                    'state': 'STARTED',
                                }
                            ]
                        }
                    }
                }
            }
        }
        shrink = Shrink(self.ilo, shrink_node=self.node_name)
        shrink.shrink_node_id = self.node_id
        self.assertRaises(ActionError, shrink._check_all_shards, testvars.named_index)



"""test_action_cold2frozen"""
# pylint: disable=attribute-defined-outside-init
from unittest import TestCase
from unittest.mock import Mock
import pytest
from curator.actions import Cold2Frozen
from curator.exceptions import CuratorException, SearchableSnapshotException
from curator import IndexList
# Get test variables and constants from a single source
from . import testvars

class TestActionCold2Frozen(TestCase):
    """TestActionCold2Frozen"""
    VERSION = {'version': {'number': '8.0.0'} }
    def builder(self):
        """Environment builder"""
        self.client = Mock()
        self.client.info.return_value = self.VERSION
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.ilo = IndexList(self.client)
    def test_init_raise_bad_index_list(self):
        """test_init_raise_bad_index_list"""
        self.assertRaises(TypeError, Cold2Frozen, 'invalid')
        with pytest.raises(TypeError):
            Cold2Frozen('not_an_IndexList')
    def test_init_add_kwargs(self):
        """test_init_add_kwargs"""
        self.builder()
        testval = {'key': 'value'}
        c2f = Cold2Frozen(self.ilo, index_settings=testval)
        assert c2f.index_settings == testval
    def test_action_generator1(self):
        """test_action_generator1"""
        self.builder()
        settings_ss   = {
            testvars.named_index: {
                'aliases': {'my_alias': {}},
                'settings': {
                    'index': {
                        'creation_date': '1456963200172',
                        'refresh_interval': '5s',
                        'lifecycle': {
                            'indexing_complete': True
                        },
                        'store': {
                            'type': 'snapshot',
                            'snapshot': {
                                'snapshot_name': 'snapname',
                                'index_name': testvars.named_index,
                                'repository_name': 'reponame',
                            }
                        }
                    }
                }
            }
        }

        self.client.indices.get_settings.return_value = settings_ss
        self.client.indices.get_alias.return_value = settings_ss
        roles = ['data_content']
        self.client.nodes.info.return_value = {'nodes': {'nodename': {'roles': roles}}}
        c2f = Cold2Frozen(self.ilo)
        snap = 'snapname'
        repo = 'reponame'
        renamed = f'partial-{testvars.named_index}'
        settings = {
            "routing": {
                "allocation": {
                    "include": {
                        "_tier_preference": roles[0]
                    }
                }
            }
        }
        expected = {
                'repository': repo, 'snapshot': snap, 'index': testvars.named_index,
                'renamed_index': renamed, 'index_settings': settings,
                'ignore_index_settings': ['index.refresh_interval'],
                'storage': 'shared_cache', 'wait_for_completion': True,
                'aliases': {'my_alias': {}}, 'current_idx': testvars.named_index
        }
        for result in c2f.action_generator():
            assert result == expected
        c2f.do_dry_run() # Do this here as it uses the same generator output.
    def test_action_generator2(self):
        """test_action_generator2"""
        self.builder()
        settings_ss   = {
            testvars.named_index: {
                'settings': {'index': {'lifecycle': {'name': 'guaranteed_fail'}}}
            }
        }
        self.client.indices.get_settings.return_value = settings_ss
        c2f = Cold2Frozen(self.ilo)
        with pytest.raises(CuratorException, match='associated with an ILM policy'):
            for result in c2f.action_generator():
                _ = result
    def test_action_generator3(self):
        """test_action_generator3"""
        self.builder()
        settings_ss   = {
            testvars.named_index: {
                'settings': {
                    'index': {
                        'lifecycle': {'indexing_complete': True},
                        'store': {'snapshot': {'partial': True}}
                    }
                }
            }
        }
        self.client.indices.get_settings.return_value = settings_ss
        c2f = Cold2Frozen(self.ilo)
        with pytest.raises(SearchableSnapshotException, match='Index is already in frozen tier'):
            for result in c2f.action_generator():
                _ = result



"""Test replica count changing functionality"""

# pylint: disable=C0115, C0116, invalid-name
import os
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestActionFileReplicas(CuratorTestCase):
    def test_increase_count(self):
        count = 2
        idx = 'my_index'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.replicas_test.format(count))
        self.create_index(idx)
        self.invoke_runner()
        assert count == int(
            self.client.indices.get_settings(index=idx)[idx]['settings']['index'][
                'number_of_replicas'
            ]
        )

    def test_no_count(self):
        self.create_index('foo')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], testvars.replicas_test.format(' '))
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_extra_option(self):
        self.create_index('foo')
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.bad_option_proto_test.format('replicas')
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code


class TestCLIReplicas(CuratorTestCase):
    def test_increase_count(self):
        count = 2
        idx = 'my_index'
        self.create_index(idx)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'replicas',
            '--count',
            str(count),
            '--filter_list',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]
        assert 0 == self.run_subprocess(
            args, logname='TestCLIOpenClosed.test_open_closed'
        )
        assert count == int(
            self.client.indices.get_settings(index=idx)[idx]['settings']['index'][
                'number_of_replicas'
            ]
        )



"""Unit tests for utils"""

from unittest import TestCase
from unittest.mock import Mock
import pytest
from curator.exceptions import (
    ActionTimeout,
    ConfigurationError,
    CuratorException,
    MissingArgument,
)
from curator.helpers.waiters import (
    health_check,
    restore_check,
    snapshot_check,
    task_check,
    wait_for_it,
)

FAKE_FAIL = Exception('Simulated Failure')


class TestHealthCheck(TestCase):
    """TestHealthCheck

    Test helpers.waiters.health_check functionality
    """

    # pylint: disable=line-too-long
    CLUSTER_HEALTH = {
        "cluster_name": "unit_test",
        "status": "green",
        "timed_out": False,
        "number_of_nodes": 7,
        "number_of_data_nodes": 3,
        "active_primary_shards": 235,
        "active_shards": 471,
        "relocating_shards": 0,
        "initializing_shards": 0,
        "unassigned_shards": 0,
        "delayed_unassigned_shards": 0,
        "number_of_pending_tasks": 0,
        "task_max_waiting_in_queue_millis": 0,
        "active_shards_percent_as_number": 100,
    }

    def test_no_kwargs(self):
        """test_no_kwargs

        Should raise a ``MissingArgument`` exception when no keyword args are passed.
        """
        client = Mock()
        with pytest.raises(
            MissingArgument, match=r'Must provide at least one keyword argument'
        ):
            health_check(client)

    def test_key_value_match(self):
        """test_key_value_match

        Should return ``True`` when matching keyword args are passed.
        """
        client = Mock()
        client.cluster.health.return_value = self.CLUSTER_HEALTH
        assert health_check(client, status='green')

    def test_key_value_no_match(self):
        """test_key_value_no_match

        Should return ``False`` when matching keyword args are passed, but no matches
        are found.
        """
        client = Mock()
        client.cluster.health.return_value = self.CLUSTER_HEALTH
        assert not health_check(client, status='red')

    def test_key_not_found(self):
        """test_key_not_found

        Should raise ``ConfigurationError`` when keyword args are passed, but keys
        match.
        """
        client = Mock()
        client.cluster.health.return_value = self.CLUSTER_HEALTH
        with pytest.raises(ConfigurationError, match=r'not in cluster health output'):
            health_check(client, foo='bar')


class TestRestoreCheck(TestCase):
    """TestRestoreCheck

    Test helpers.waiters.restore_check functionality
    """

    SNAP_NAME = 'snap_name'
    NAMED_INDICES = ["index-2015.01.01", "index-2015.02.01"]

    def test_fail_to_get_recovery(self):
        """test_fail_to_get_recovery

        Should raise ``CuratorException`` when an upstream Exception is encountered
        """
        client = Mock()
        client.indices.recovery.side_effect = FAKE_FAIL
        with pytest.raises(
            CuratorException, match=r'Unable to obtain recovery information'
        ):
            restore_check(client, [])

    def test_incomplete_recovery(self):
        """test_incomplete_recovery

        Should return ``False`` when recovery is incomplete
        """
        client = Mock()
        # :pylint disable=line-too-long
        client.indices.recovery.return_value = {
            'index-2015.01.01': {'shards': [{'stage': 'INDEX'}]},
            'index-2015.02.01': {'shards': [{'stage': 'INDEX'}]},
        }
        assert not restore_check(client, self.NAMED_INDICES)

    def test_completed_recovery(self):
        """test_completed_recovery

        Should return ``True`` when recovery is complete
        """
        client = Mock()
        # :pylint disable=line-too-long
        client.indices.recovery.return_value = {
            'index-2015.01.01': {'shards': [{'stage': 'DONE'}]},
            'index-2015.02.01': {'shards': [{'stage': 'DONE'}]},
        }
        assert restore_check(client, self.NAMED_INDICES)

    def test_empty_recovery(self):
        """test_empty_recovery

        Should return ``False`` when an empty response comes back
        """
        client = Mock()
        client.indices.recovery.return_value = {}
        assert not restore_check(client, self.NAMED_INDICES)


class TestSnapshotCheck(TestCase):
    """TestSnapshotCheck

    Test helpers.waiters.snapshot_check functionality
    """

    # :pylint disable=line-too-long
    SNAP_NAME = 'snap_name'
    NAMED_INDICES = ["index-2015.01.01", "index-2015.02.01"]

    def test_fail_to_get_snapshot(self):
        """test_fail_to_get_snapshot

        Should raise ``CuratorException`` when another upstream Exception occurs.
        """
        client = Mock()
        client.snapshot.get.side_effect = FAKE_FAIL
        self.assertRaises(CuratorException, snapshot_check, client)

    def test_in_progress(self):
        """test_in_progress

        Should return ``False`` when state is ``IN_PROGRESS``.
        """
        client = Mock()
        test_val = {
            'snapshots': [
                {
                    'state': 'IN_PROGRESS',
                    'snapshot': self.SNAP_NAME,
                    'indices': self.NAMED_INDICES,
                }
            ]
        }
        client.snapshot.get.return_value = test_val
        assert not snapshot_check(client, repository='foo', snapshot=self.SNAP_NAME)

    def test_success(self):
        """test_success

        Should return ``True`` when state is ``SUCCESS``.
        """
        client = Mock()
        test_val = {
            'snapshots': [
                {
                    'state': 'SUCCESS',
                    'snapshot': self.SNAP_NAME,
                    'indices': self.NAMED_INDICES,
                }
            ]
        }
        client.snapshot.get.return_value = test_val
        assert snapshot_check(client, repository='foo', snapshot=self.SNAP_NAME)

    def test_partial(self):
        """test_partial

        Should return ``True`` when state is ``PARTIAL``.
        """
        client = Mock()
        test_val = {
            'snapshots': [
                {
                    'state': 'PARTIAL',
                    'snapshot': self.SNAP_NAME,
                    'indices': self.NAMED_INDICES,
                }
            ]
        }
        client.snapshot.get.return_value = test_val
        assert snapshot_check(client, repository='foo', snapshot=self.SNAP_NAME)

    def test_failed(self):
        """test_failed

        Should return ``True`` when state is ``FAILED``.
        """
        client = Mock()
        test_val = {
            'snapshots': [
                {
                    'state': 'FAILED',
                    'snapshot': self.SNAP_NAME,
                    'indices': self.NAMED_INDICES,
                }
            ]
        }
        client.snapshot.get.return_value = test_val
        assert snapshot_check(client, repository='foo', snapshot=self.SNAP_NAME)

    def test_other(self):
        """test_other

        Should return ``True`` when state is anything other than ``IN_PROGRESS`` or the
        above.
        """
        client = Mock()
        test_val = {
            'snapshots': [
                {
                    'state': 'SOMETHINGELSE',
                    'snapshot': self.SNAP_NAME,
                    'indices': self.NAMED_INDICES,
                }
            ]
        }
        client.snapshot.get.return_value = test_val
        assert snapshot_check(client, repository='foo', snapshot=self.SNAP_NAME)


class TestTaskCheck(TestCase):
    """TestTaskCheck

    Test helpers.waiters.task_check functionality
    """

    # pylint: disable=line-too-long
    PROTO_TASK = {
        'node': 'I0ekFjMhSPCQz7FUs1zJOg',
        'description': 'UNIT TEST',
        'running_time_in_nanos': 1637039537721,
        'action': 'indices:data/write/reindex',
        'id': 54510686,
        'start_time_in_millis': 1489695981997,
    }
    GENERIC_TASK = {'task': 'I0ekFjMhSPCQz7FUs1zJOg:54510686'}

    def test_bad_task_id(self):
        """test_bad_task_id

        Should raise ``CuratorException`` if a bad value for ``task_id`` is passed
        """
        client = Mock()
        client.tasks.get.side_effect = FAKE_FAIL
        with pytest.raises(
            CuratorException, match=r'Unable to obtain task information for task'
        ):
            task_check(client, 'foo')

    def test_incomplete_task(self):
        """test_incomplete_task

        Should return ``False`` if task is incomplete
        """
        client = Mock()
        test_task = {
            'completed': False,
            'task': self.PROTO_TASK,
            'response': {'failures': []},
        }
        client.tasks.get.return_value = test_task
        assert not task_check(client, task_id=self.GENERIC_TASK['task'])

    def test_complete_task(self):
        """test_complete_task

        Should return ``True`` if task is complete
        """
        client = Mock()
        test_task = {
            'completed': True,
            'task': self.PROTO_TASK,
            'response': {'failures': []},
        }
        client.tasks.get.return_value = test_task
        assert task_check(client, task_id=self.GENERIC_TASK['task'])


class TestWaitForIt(TestCase):
    """TestWaitForIt

    Test helpers.waiters.wait_for_it functionality
    """

    # pylint: disable=line-too-long
    def test_bad_action(self):
        """test_bad_action

        Should raise a ``ConfigurationError`` exception if ``action`` is invalid
        """
        client = Mock()
        # self.assertRaises(ConfigurationError, wait_for_it, client, 'foo')
        with pytest.raises(ConfigurationError, match=r'"action" must be one of'):
            wait_for_it(client, 'foo')

    def test_reindex_action_no_task_id(self):
        """test_reindex_action_no_task_id

        Should raise a ``MissingArgument`` exception if ``task_id`` is missing for
        ``reindex``
        """
        client = Mock()
        # self.assertRaises(MissingArgument, wait_for_it, client, 'reindex')
        with pytest.raises(MissingArgument, match=r'A task_id must accompany "action"'):
            wait_for_it(client, 'reindex')

    def test_snapshot_action_no_snapshot(self):
        """test_snapshot_action_no_snapshot

        Should raise a ``MissingArgument`` exception if ``snapshot`` is missing for
        ``snapshot``
        """
        client = Mock()
        # self.assertRaises(MissingArgument, wait_for_it, client,
        #   'snapshot', repository='foo')
        with pytest.raises(
            MissingArgument, match=r'A snapshot and repository must accompany "action"'
        ):
            wait_for_it(client, 'snapshot', repository='foo')

    def test_snapshot_action_no_repository(self):
        """test_snapshot_action_no_repository

        Should raise a ``MissingArgument`` exception if ``repository`` is missing for
        ``snapshot``
        """
        client = Mock()
        # self.assertRaises(MissingArgument, wait_for_it, client,
        #   'snapshot', snapshot='foo')
        with pytest.raises(
            MissingArgument, match=r'A snapshot and repository must accompany "action"'
        ):
            wait_for_it(client, 'snapshot', snapshot='foo')

    def test_restore_action_no_indexlist(self):
        """test_restore_action_no_indexlist

        Should raise a ``MissingArgument`` exception if ``index_list`` is missing for
        ``restore``
        """
        client = Mock()
        # self.assertRaises(MissingArgument, wait_for_it, client, 'restore')
        with pytest.raises(
            MissingArgument, match=r'An index_list must accompany "action"'
        ):
            wait_for_it(client, 'restore')

    def test_reindex_action_bad_task_id(self):
        """test_reindex_action_bad_task_id

        Should raise a ``CuratorException`` exception if there's a bad task_id

        This is kind of a fake fail, even in the code.
        """
        client = Mock()
        client.tasks.get.return_value = {'a': 'b'}
        client.tasks.get.side_effect = FAKE_FAIL
        # self.assertRaises(CuratorException, wait_for_it,
        #       client, 'reindex', task_id='foo')
        with pytest.raises(CuratorException, match=r'Unable to find task_id'):
            wait_for_it(client, 'reindex', task_id='foo')

    def test_reached_max_wait(self):
        """test_reached_max_wait

        Should raise a ``ActionTimeout`` exception if we've waited past the defined
        timeout period
        """
        client = Mock()
        client.cluster.health.return_value = {'status': 'red'}
        # self.assertRaises(ActionTimeout, wait_for_it, client, 'replicas',
        # wait_interval=1, max_wait=1)
        with pytest.raises(
            ActionTimeout, match=r'failed to complete in the max_wait period'
        ):
            wait_for_it(client, 'replicas', wait_interval=1, max_wait=1)
